(*****************************************************************)
(*                                                               *)
(*              3D Rendering Engine Camera class                 *)
(*                                                               *)
(*                   Written by Peter Bone                       *)
(*                   peterbone@hotmail.com                       *)
(*           http://www.geocities.com/peter_bone_uk              *)
(*                                                               *)
(*                Last Modified   5 / 7 / 2005                   *)
(*                                                               *)
(*    This unit contains the TCamera class used to store and     *)
(*    modify the properties of the camera including position,    *)
(*    orientation, the viewing frustum, the projection plane     *)
(*                                                               *)
(*       This code is free for use but please give credit        *)
(*                                                               *)
(*****************************************************************)

unit Camera;

interface

uses
  Global, Windows, Math, Classes;

type

  // an infinite plane
  TPlane = record
    Normal : TPoint3D; // unit vector normal to plane
    Distance : Real; // distance to plane from origin along normal
  end;

  TCamera = class(TObject)
    private

      function InverseLookAt(AVector : TPoint3D) : TPoint3D;

    public

      BaseFrustum, Frustum : array[0..5] of TPlane;
      ScreenMid : TPoint;
      PerspectiveOffset : integer;
      Position : TPoint3D; // position of camera in world
      Ahead : TPoint3D; // forward direction - must be unit
      Down  : TPoint3D; // up direction - must be unit
      Right : TPoint3D; // right direction - must be unit
      NearPlaneOffset : Real;
      FarPlaneOffset : Real;

      procedure Assign(const Source : TCamera);
      procedure ComputeFrustum;
      procedure PositionFrustum;
      function PointInFrustum(APoint : TPoint3D) : boolean;
      function SphereInFrustum(ASphere : TSphere) : byte;
      procedure LookAt(var APoint : TPoint3D);
      procedure Rotate(AAngle : Real ; AAxis : TAxis);
      procedure Move(ADisp : Real ; AAxis : TAxis);
      function PointPlaneSide(AFrustumPlane : byte ; APoint : TPoint3D) : boolean;
      procedure SaveToStream(AStream : TMemoryStream);
      procedure LoadFromStream(AStream : TMemoryStream);

      constructor Create;

    end;

Const
  // sphere in frustum states
  Inside = 0;
  Outside = 1;
  Intersecting = 2;

implementation

constructor TCamera.Create;
begin
  Position := Origin;
  Ahead := Point3D(0, 0, 1);
  Down  := Point3D(0, 1, 0);
  Right := Point3D(1, 0, 0);
  NearPlaneOffset := 10;
  FarPlaneOffset := 10000;
end;

procedure TCamera.Assign(const Source : TCamera);
begin
  ScreenMid := Source.ScreenMid;
  PerspectiveOffset := Source.PerspectiveOffset;
  Position := Source.Position;
  Ahead := Source.Ahead;
  Down := Source.Down;
  Right := Source.Right;
  NearPlaneOffset := Source.NearPlaneOffset;
  System.Move(Source.BaseFrustum[0], BaseFrustum[0], SizeOf(TPlane)*6);
end;

// compute the plane normals of the base frustum - frustum at origin
procedure TCamera.ComputeFrustum;
Var
  LProjectionPlaneCorners : array[0..3] of TPoint3D;
begin
  // calculate coordinates of projection plane corners - if camera at origin
  LProjectionPlaneCorners[0] := Point3D( ScreenMid.X, ScreenMid.Y, PerspectiveOffset);
  LProjectionPlaneCorners[1] := Point3D(-ScreenMid.X, ScreenMid.Y, PerspectiveOffset);
  LProjectionPlaneCorners[2] := Point3D(-ScreenMid.X,-ScreenMid.Y, PerspectiveOffset);
  LProjectionPlaneCorners[3] := Point3D( ScreenMid.X,-ScreenMid.Y, PerspectiveOffset);

  BaseFrustum[1].Normal := Point3D(0, 0, -1); // far plane

  BaseFrustum[2].Normal := UnitVector( CrossProduct(LProjectionPlaneCorners[3],
                                       LProjectionPlaneCorners[0]) ); // right
  BaseFrustum[3].Normal := UnitVector( CrossProduct(LProjectionPlaneCorners[0],
                                       LProjectionPlaneCorners[1]) ); // bottom
  BaseFrustum[4].Normal := UnitVector( CrossProduct(LProjectionPlaneCorners[1],
                                       LProjectionPlaneCorners[2]) ); // left
  BaseFrustum[5].Normal := UnitVector( CrossProduct(LProjectionPlaneCorners[2],
                                       LProjectionPlaneCorners[3]) ); // top
end;

// compute frustum by moving the base frustum to the cameras position
procedure TCamera.PositionFrustum;
Var
  LPlane : byte;
begin
  // rotate plane normals to align with camera orientation
  for LPlane := 1 to 5 do
    Frustum[LPlane].Normal := InverseLookAt(BaseFrustum[LPlane].Normal);

  Frustum[0].Normal := OpositeDirection(Frustum[1].Normal); // near plane

  // calculate plane distances using D = n.p (n = normal, p = point on plane)
  Frustum[0].Distance := DotProduct(Frustum[0].Normal, Add(Position, ScalarProduct(Ahead, NearPlaneOffset)));
  Frustum[1].Distance := DotProduct(Frustum[1].Normal, Add(Position, ScalarProduct(Ahead, FarPlaneOffset)));
  for LPlane := 2 to 5 do
    Frustum[LPlane].Distance := DotProduct(Frustum[LPlane].Normal, Position);
end;

// transform a 3D point from world space to view space
procedure TCamera.LookAt(var APoint : TPoint3D);
Var
  TempPoint : TPoint3D;
begin
  TempPoint.X := DotProduct(APoint, Right);
  TempPoint.Y := DotProduct(APoint, Down);
  TempPoint.Z := DotProduct(APoint, Ahead);
  APoint := TempPoint;
end;

// rotate a vector from view space to world space
function TCamera.InverseLookAt(AVector : TPoint3D) : TPoint3D;
begin
  Result.X := AVector.X * Right.X + AVector.Y * Down.X - AVector.Z * Ahead.X;
  Result.Y := AVector.X * Right.Y + AVector.Y * Down.Y - AVector.Z * Ahead.Y;
  Result.Z := AVector.X * Right.Z + AVector.Y * Down.Z - AVector.Z * Ahead.Z;
end;

// check if a point is inside the cameras field of vision
// inside or outside frustum
function TCamera.PointInFrustum(APoint : TPoint3D) : boolean;
Var
  LPlane : byte;
begin
  for LPlane := 0 to 5 do begin
    if not PointPlaneSide(LPlane, APoint) then begin
      Result := False;
      Exit; // outside frustum - return false
    end;
  end;

  Result := True; // inside frustum
end;

// determine if a point is on the postive or negative side of a frustum plane
function TCamera.PointPlaneSide(AFrustumPlane : byte ; APoint : TPoint3D) : boolean;
begin
  if DotProduct(Frustum[AFrustumPlane].Normal, APoint) >
     Frustum[AFrustumPlane].Distance then Result := False
                                     else Result := True;
end;

// check if a sphere is inside the cameras field of vision
// inside, outside or intersecting frustum
function TCamera.SphereInFrustum(ASphere : TSphere) : byte;
Var
  LPlane : byte;
  LCenterDist : Real;
begin
  Result := Inside;

  for LPlane := 0 to 5 do begin
    LCenterDist := Frustum[LPlane].Distance -
                   DotProduct(Frustum[LPlane].Normal, ASphere.Center);
    if LCenterDist < -ASphere.Radius then begin
      Result := Outside; // outside frustum
      Exit;
    end;
    if (LCenterDist <= ASphere.Radius) and (LCenterDist >= -ASphere.Radius) then
      Result := Intersecting; // intersecting frustum - may still be outside other planes
  end;
end;

// rotate around the axis of the camera - AxisX=pitch, AxisY=yaw, AxisZ=roll
procedure TCamera.Rotate(AAngle : Real ; AAxis : TAxis);
Var
  Lsin, Lcos : Extended;
  Temp : TPoint3D;
begin
  SinCos(AAngle, Lsin, Lcos);

  if AAxis = AxisX then begin // rotate about cameras right vector - pitch
    Temp := Add(ScalarProduct(Down, Lcos), ScalarProduct(Ahead, Lsin));
    AHead := Subtract(ScalarProduct(Ahead, Lcos), ScalarProduct(Down, Lsin));
    Down := Temp;
  end else if AAxis = AxisY then begin // rotate about cameras down vector - yaw
    Temp := Add(ScalarProduct(Ahead, Lcos), ScalarProduct(Right, Lsin));
    Right := Subtract(ScalarProduct(Right, Lcos), ScalarProduct(Ahead, Lsin));
    Ahead := Temp;
  end else begin // rotate about cameras ahead vector - roll
    Temp := Add(ScalarProduct(Right, Lcos), ScalarProduct(Down, Lsin));
    Down := Subtract(ScalarProduct(Down, Lcos), ScalarProduct(Right, Lsin));
    Right := Temp;
  end;
end;

// move in direction of a given camera axis
procedure TCamera.Move(ADisp : Real ; AAxis : TAxis);
begin
  Case AAxis of
    AxisX : AddP( Position, ScalarProduct(Right, ADisp) ); // move right or left
    AxisY : AddP( Position, ScalarProduct(Down , ADisp) ); // move down or up
    AxisZ : AddP( Position, ScalarProduct(Ahead, ADisp) ); // move forward or back
  end;    
end;

// save the camera data to a memory stream
procedure TCamera.SaveToStream(AStream : TMemoryStream);
begin
  AStream.Write(ScreenMid, SizeOf(TPoint));
  AStream.Write(PerspectiveOffset, SizeOf(Integer));
  AStream.Write(Position, SizeOf(TPoint3D));
  AStream.Write(Ahead, SizeOf(TPoint3D));
  AStream.Write(Down, SizeOf(TPoint3D));
  AStream.Write(Right, SizeOf(TPoint3D));
  AStream.Write(NearPlaneOffset, SizeOf(Real));
  AStream.Write(FarPlaneOffset, SizeOf(Real));
end;

// load the camera data from a memory stream
procedure TCamera.LoadFromStream(AStream : TMemoryStream);
begin
  AStream.Read(ScreenMid, SizeOf(TPoint));
  AStream.Read(PerspectiveOffset, SizeOf(Integer));
  AStream.Read(Position, SizeOf(TPoint3D));
  AStream.Read(Ahead, SizeOf(TPoint3D));
  AStream.Read(Down, SizeOf(TPoint3D));
  AStream.Read(Right, SizeOf(TPoint3D));
  AStream.Read(NearPlaneOffset, SizeOf(Real));
  AStream.Read(FarPlaneOffset, SizeOf(Real));
  ComputeFrustum;
end;

end.
