unit Global;

interface

uses
  Graphics, ColorConv, Dialogs, Classes, ZLib;

type

  TColorArray = array of TColor;

  TPointFloat = record
    X : single;
    Y : single;
  end;

  TPoint3D = record
    X : Real;
    Y : Real;
    Z : Real;
  end;

  TPoint3DSingle = record
    X : Single;
    Y : Single;
    Z : Single;
  end;

  TPoints = array of TPoint3D;

  TVertex = record
    Point : TPoint3D; // coordinate of vertex in world space
    ViewPoint : TPoint3D; // coordinate of vertex in view space
    Normal : TPoint3D; // vector normal to surface at vertex
    Visible : boolean; // flag to show if vertex is part of a visible face
    Lum : byte; // Luminosity at vertex for gouraud polygon filling
    ScreenPoint : TPointFloat; // perspective mapping of vertex on screen
  end;

  TEdge = record
    Point1 : integer;
    Point2 : integer;
  end;

  TFace = record
    Corners : array of integer; // indexes the vertices that make up the face
    Center : TPoint3D; // position of center of face in world space
    ViewCenter : TPoint3D; // position of center of face in view space
//    Distance : Real; // distance to face center from viewer
    Normal : TPoint3D; // normal vector to the plane of the face
    HLS : THLS;
    Curved : boolean; // indicates whether face is curved or flat
  end;
  TFaceArray = array of TFace;

  TAxis = (AxisX, AxisY, AxisZ);

  TSphere = record
    Center : TPoint3D;
    Radius : Real;
  end;  

procedure AssignFace(const Source : TFace ; out Dest : TFace);
procedure RotateVector(var AVector : TPoint3D ; Asin, Acos : Real ; AAxis : TAxis);
procedure RotatePoint(var AVector : TPoint3D ; APivot : TPoint3D ; Asin, Acos : Real ; AAxis : TAxis);
function CrossProduct(AVector1, AVector2 : TPoint3D) : TPoint3D;
function DotProduct(AVector1, AVector2 : TPoint3D) : Real;
function DotProductS(AVector1, AVector2 : TPoint3DSingle) : Single;
function ScalarProduct(AVector : TPoint3D ; AScalar : Real) : TPoint3D;
function ScalarProductS(AVector : TPoint3DSingle ; AScalar : Single) : TPoint3DSingle;
procedure ScalarProductP(var AVector : TPoint3D ; AScalar : Real);
function UnitVector(AVector : TPoint3D) : TPoint3D;
function UnitVectorS(AVector : TPoint3DSingle) : TPoint3DSingle;
procedure UnitVectorP(var AVector : TPoint3D);
function Modulus(AVector : TPoint3D) : Real;
function ModulusS(AVector : TPoint3DSingle) : Single;
function Add(AVec1, AVec2 : TPoint3D) : TPoint3D;
procedure AddP(var AVec1 : TPoint3D ; AVec2 : TPoint3D);
function Subtract(AVec1, AVec2 : TPoint3D) : TPoint3D;
procedure SubtractP(var AVec1 : TPoint3D ; AVec2 : TPoint3D);
function MidPoint(APoint1, APoint2 : TPoint3D) : TPoint3D;
function OpositeDirection(AVector : TPoint3D) : TPoint3D;
function Point3D(AX, AY, AZ : Real) : TPoint3D;
function Point2D(AX, AY : Single) : TPointFloat;
function CalcAngle(APoint1, APoint2 : TPointFloat) : real;
function Point3DSingle(AX, AY, AZ : Single) : TPoint3DSingle;
function Point3DDtoS(P : TPoint3D) : TPoint3DSingle;

procedure PackStream(inStream, outStream :TStream;
                     CompressionLevel:TCompressionLevel);
procedure ExpandStream(inStream, outStream :TStream);

Var
  Origin : TPoint3D;

implementation

procedure AssignFace(const Source : TFace ; out Dest : TFace);
Var
  LLen : integer;
begin
  with Dest do begin
    LLen := length(Source.Corners);
    SetLength(Corners, LLen);
    Move(Source.Corners[0], Corners[0], SizeOf(integer)*LLen);
    Center := Source.Center;
    ViewCenter := Source.ViewCenter;
    Normal := Source.Normal;
    HLS := Source.HLS;
    Curved := Source.Curved;
  end;
end;

// vector handling routines

// rotate a vector by a given angle and around a given axis
// pass in sin and cos of the angle for speed when rotating multiple vectors by
// the same angle - only have to calculate sin and cos once
procedure RotateVector(var AVector : TPoint3D ; Asin, Acos : Real ; AAxis : TAxis);
Var
  TempVector : TPoint3D;
begin
  if AAxis = AxisX then begin
    TempVector.X := AVector.X;
    TempVector.Y := AVector.Y * Acos - AVector.Z * Asin;
    TempVector.Z := AVector.Y * Asin + AVector.Z * Acos;
  end else if AAxis = AxisY then begin
    TempVector.X := AVector.X * Acos + AVector.Z * Asin;
    TempVector.Y := AVector.Y;
    TempVector.Z := AVector.Z * Acos - AVector.X * Asin;
  end else begin
    TempVector.X := AVector.X * Acos - AVector.Y * Asin;
    TempVector.Y := AVector.X * Asin + AVector.Y * Acos;
    TempVector.Z := AVector.Z;
  end;

  AVector := TempVector;
end;

// rotate a point around a pivot point by a given angle and around a given axis
procedure RotatePoint(var AVector : TPoint3D ; APivot : TPoint3D ; Asin, Acos : Real ; AAxis : TAxis);
Var
  TempVector, LDiff : TPoint3D;
begin
  LDiff := Subtract(AVector, APivot);

  if AAxis = AxisX then begin
    TempVector.X := AVector.X;
    TempVector.Y := LDiff.Y * Acos - LDiff.Z * Asin + APivot.Y;
    TempVector.Z := LDiff.Y * Asin + LDiff.Z * Acos + APivot.Z;
  end else if AAxis = AxisY then begin
    TempVector.X := LDiff.X * Acos + LDiff.Z * Asin + APivot.X;
    TempVector.Y := AVector.Y;
    TempVector.Z := LDiff.Z * Acos - LDiff.X * Asin + APivot.Z;
  end else begin
    TempVector.X := LDiff.X * Acos - LDiff.Y * Asin + APivot.X;
    TempVector.Y := LDiff.X * Asin + LDiff.Y * Acos + APivot.Y;
    TempVector.Z := AVector.Z;
  end;

  AVector := TempVector;
end;

// calculate the vector normal to 2 vectors
function CrossProduct(AVector1, AVector2 : TPoint3D) : TPoint3D;
begin
  Result.X := AVector1.Y * AVector2.Z - AVector1.Z * AVector2.Y;
  Result.Y := AVector1.Z * AVector2.X - AVector1.X * AVector2.Z;
  Result.Z := AVector1.X * AVector2.Y - AVector1.Y * AVector2.X;
end;

// calculate the dot product of 2 vectors
function DotProduct(AVector1, AVector2 : TPoint3D) : Real;
begin
  Result := AVector1.X * AVector2.X +
            AVector1.Y * AVector2.Y +
            AVector1.Z * AVector2.Z;
end;

function DotProductS(AVector1, AVector2 : TPoint3DSingle) : Single;
begin
  Result := AVector1.X * AVector2.X +
            AVector1.Y * AVector2.Y +
            AVector1.Z * AVector2.Z;
end;

// multiply a vector by a scalar
function ScalarProduct(AVector : TPoint3D ; AScalar : Real) : TPoint3D;
begin
  Result := Point3D(AVector.X * AScalar, AVector.Y * AScalar, AVector.Z * AScalar);
end;

function ScalarProductS(AVector : TPoint3DSingle ; AScalar : Single) : TPoint3DSingle;
begin
  Result.X := AVector.X * AScalar;
  Result := Point3DSingle(AVector.X * AScalar, AVector.Y * AScalar, AVector.Z * AScalar);
end;

// multiply a vector by a scalar and store result in the vector
// using a procedure and returning into a var is faster and reduces code
procedure ScalarProductP(var AVector : TPoint3D ; AScalar : Real);
begin
  AVector.X := AVector.X * AScalar;
  AVector.Y := AVector.Y * AScalar;
  AVector.Z := AVector.Z * AScalar;
end;

// return the unit vector of a vector
function UnitVector(AVector : TPoint3D) : TPoint3D;
begin
  try

    Result := ScalarProduct(AVector, 1 / Modulus(AVector));

  except
    ShowMessage('Exception in UnitVector Method');
  end;
end;

function UnitVectorS(AVector : TPoint3DSingle) : TPoint3DSingle;
begin
  try

    Result := ScalarProductS(AVector, 1 / ModulusS(AVector));

  except
    ShowMessage('Exception in UnitVectorS Method');
  end;
end;

// reduce a vector to a unit vector
procedure UnitVectorP(var AVector : TPoint3D);
Var
  LModulusInv : Real;
begin
  try

    LModulusInv := 1 / Modulus(AVector); // * is faster than /
    ScalarProductP(AVector, LModulusInv);

  except
    ShowMessage('Exception in UnitVectorP Method');
  end;
end;

// calculate the modulus (length) of a vector
function Modulus(AVector : TPoint3D) : Real;
begin
  Result := Sqrt( Sqr(AVector.X) + Sqr(AVector.Y) + Sqr(AVector.Z) );
end;

function ModulusS(AVector : TPoint3DSingle) : Single;
begin
  Result := Sqrt( Sqr(AVector.X) + Sqr(AVector.Y) + Sqr(AVector.Z) );
end;

// add two vectors together
function Add(AVec1, AVec2 : TPoint3D) : TPoint3D;
begin
  Result := Point3D(AVec1.X + AVec2.X, AVec1.Y + AVec2.Y, AVec1.Z + AVec2.Z);
end;

// add two vectors and store result in the first vector
procedure AddP(var AVec1 : TPoint3D ; AVec2 : TPoint3D);
begin
  AVec1.X := AVec1.X + AVec2.X;
  AVec1.Y := AVec1.Y + AVec2.Y;
  AVec1.Z := AVec1.Z + AVec2.Z;
end;

// subtract one vector from another
function Subtract(AVec1, AVec2 : TPoint3D) : TPoint3D;
begin
  Result := Point3D(AVec1.X - AVec2.X, AVec1.Y - AVec2.Y, AVec1.Z - AVec2.Z);
end;

// subtract vector2 from vector1 and store result in vector1
procedure SubtractP(var AVec1 : TPoint3D ; AVec2 : TPoint3D);
begin
  AVec1.X := AVec1.X - AVec2.X;
  AVec1.Y := AVec1.Y - AVec2.Y;
  AVec1.Z := AVec1.Z - AVec2.Z;
end;

// calculate the mid-point of 2 coordinates
function MidPoint(APoint1, APoint2 : TPoint3D) : TPoint3D;
begin
  Result.X := (APoint1.X + APoint2.X) * 0.5;
  Result.Y := (APoint1.Y + APoint2.Y) * 0.5;
  Result.Z := (APoint1.Z + APoint2.Z) * 0.5;
end;

// reverse the direction of a vector
function OpositeDirection(AVector : TPoint3D) : TPoint3D;
begin
  Result.X := -AVector.X;
  Result.Y := -AVector.Y;
  Result.Z := -AVector.Z;
end;

// TPoint3D type-cast
function Point3D(AX, AY, AZ : Real) : TPoint3D;
begin
  Result.X := AX;
  Result.Y := AY;
  Result.Z := AZ;
end;

// TPoint3DSingle type-cast
function Point3DSingle(AX, AY, AZ : Single) : TPoint3DSingle;
begin
  Result.X := AX;
  Result.Y := AY;
  Result.Z := AZ;
end;

// TPointFloat type-cast
function Point2D(AX, AY : Single) : TPointFloat;
begin
  Result.X := AX;
  Result.Y := AY;
end;

// TPoint3D to TPoint3DSingle conversion
function Point3DDtoS(P : TPoint3D) : TPoint3DSingle;
begin
  Result.X := P.X;
  Result.Y := P.Y;
  Result.Z := P.Z;
end;

// calculate the polar angle of a vector
function CalcAngle(APoint1, APoint2 : TPointFloat) : real;
begin
  // check to prevent EZeroDivide
  if APoint1.Y = APoint2.Y then begin
    if APoint1.X < APoint2.X then Result := 0
                             else Result := pi;
    Exit;
  end;

  Result := Arctan(Abs((APoint1.X-APoint2.X)/(APoint1.Y-APoint2.Y)));
  if APoint2.Y > APoint1.Y then begin
    if APoint2.X >= APoint1.X then Result := pi*0.5 - Result
                              else Result := Result + pi*0.5;
  end else begin
    if APoint2.X >= APoint1.X then Result := Result + pi*1.5
                              else Result := pi*1.5 - Result;
  end;
end;

// compress a stream using zlib
procedure PackStream(inStream, outStream :TStream;
                     CompressionLevel:TCompressionLevel);
var ZStream: TCompressionStream; 
begin 
    ZStream := TCompressionStream.Create(CompressionLevel, OutStream); 
  try 
    ZStream.CopyFrom(InStream, 0); 
  finally 
    ZStream.Free; 
  end; 
end; 

// decompress a stream using zlib
procedure ExpandStream(inStream, outStream :TStream);
const
  BufferSize = 4096;
var 
  Count: Integer;
  ZStream: TDecompressionStream;
  Buffer: array[0..BufferSize-1] of Byte; 
begin
  ZStream := TDecompressionStream.Create(InStream); 
  try 
    while True do 
      begin 
        Count := ZStream.Read(Buffer, BufferSize); 
        if Count <> 0 then OutStream.WriteBuffer(Buffer, Count) else Break; 
      end; 
  finally 
    ZStream.Free; 
  end; 
end;

end.
