(*****************************************************************)
(*                                                               *)
(*              3D Rendering Engine Model class                  *)
(*                                                               *)
(*                   Written by Peter Bone                       *)
(*                   peterbone@hotmail.com                       *)
(*           http://www.geocities.com/peter_bone_uk              *)
(*                                                               *)
(*                Last Modified  5 / 7 / 2005                    *)
(*                                                               *)
(*    This unit contains the T3DModel class used to store the    *)
(*       properties of a 3D object and transform the model       *)
(*                                                               *)
(*       This code is free for use but please give credit        *)
(*                                                               *)
(*****************************************************************)

unit Model3D;

interface

uses
  Math, Dialogs, Classes, ColorConv, Global, ZLib;

type

  T3DModel = Class(TObject)
    private

    public

      Vertices : array of TVertex;
      Edges : array of TEdge;
      Faces : TFaceArray;
      BoundingSphere : TSphere;
      Center : TPoint3D;
      Color : TRGB; // for wireframe only

      procedure Assign(const Source : T3DModel);
      procedure Translate(ADisp : TPoint3D);
      procedure Rotate(AAngle : Real ; AAxis : TAxis);
      procedure Resize(ARatio : Real);
      procedure SaveToStream(AStream : TMemoryStream);
      procedure LoadFromStream(AStream : TMemoryStream);
      procedure SaveToFile(APath : String);
      procedure LoadFromFile(APath : String);

      procedure ComputeFaceNormals;
      procedure ComputeVertexNormals;
      procedure AssignEdges;
      procedure ComputeFaceCenters;
      procedure ComputeBoundingSphere;

      constructor Create;

    end;

implementation

constructor T3DModel.Create;
begin
  // set class default parameters
  Center := Origin;
end;

// assign from another model
procedure T3DModel.Assign(const Source : T3DModel);
Var
  LLen, LFace : integer;
begin
  try

    // assign vertices
    LLen := length(Source.Vertices);
    SetLength(Vertices, LLen);
    if LLen > 0 then Move(Source.Vertices[0], Vertices[0], SizeOf(TVertex)*LLen);
    // assign edges
    LLen := length(Source.Edges);
    SetLength(Edges, LLen);
    if LLen > 0 then Move(Source.Edges[0], Edges[0], SizeOf(TEdge)*LLen);
    // assign faces
    LLen := length(Source.Faces);
    SetLength(Faces, LLen);
    if LLen > 0 then for LFace := 0 to LLen - 1 do
      AssignFace(Source.Faces[LFace], Faces[LFace]);
    // assign bounding sphere
    BoundingSphere := Source.BoundingSphere;
    // assign center
    Center := Source.Center;
    // assign color
    Color := Source.Color;

  except
    ShowMessage('Exception in Model Assign Method');
  end;
end;

// translate an object
procedure T3DModel.Translate(ADisp : TPoint3D);
Var
  LIndex : integer;
begin
  try

    for LIndex := 0 to High(Vertices) do AddP( Vertices[LIndex].Point, ADisp );

    for LIndex := 0 to High(Faces) do AddP( Faces[LIndex].Center, ADisp );

    AddP( Center, ADisp );
    AddP( BoundingSphere.Center, ADisp );

  except
    ShowMessage('Exception in Model Translate Method');
  end;
end;

// rotate an object around a given axis
procedure T3DModel.Rotate(AAngle : Real ; AAxis : TAxis);
Var
  LIndex : integer;
  Lsin, Lcos : Extended;
begin
  try

    SinCos(AAngle, Lsin, Lcos);

    for LIndex := 0 to High(Vertices) do begin
      // vertex points
      RotatePoint(Vertices[LIndex].Point, Center, Lsin, Lcos, AAxis);
      // vertex normals
      RotateVector(Vertices[LIndex].Normal, Lsin, Lcos, AAxis);
    end;
    for LIndex := 0 to High(Faces) do begin
      // face normals
      RotateVector(Faces[LIndex].Normal, Lsin, Lcos, AAxis);
      // face centers
      RotatePoint(Faces[LIndex].Center, Center, Lsin, Lcos, AAxis);
    end;

  except
    ShowMessage('Exception in Model Rotate Method');
  end;
end;

// resize an object by a given ratio keeping it's center in the same place
procedure T3DModel.Resize(ARatio : Real);
Var
  LOffset : TPoint3D;
  Lindex : integer;
begin
  try

    // move to origin
    LOffset := Center;
    Translate(OpositeDirection(LOffset));

    // re-scale vertex co-ordinates
    for Lindex := 0 to High(Vertices) do
      Vertices[LIndex].Point := ScalarProduct(Vertices[LIndex].Point, ARatio);

    // re-scale face centers
    for Lindex := 0 to High(Faces) do
      Faces[Lindex].Center := ScalarProduct(Faces[Lindex].Center, ARatio);

    // re-scale bounding sphere
    BoundingSphere.Center := ScalarProduct(BoundingSphere.Center, ARatio);
    BoundingSphere.Radius := BoundingSphere.Radius * ARatio;

    // move back to original position  
    Translate(LOffset);

  except
    ShowMessage('Exception in Model Resize Method');
  end;
end;

// calculate the normal vector of a face
procedure T3DModel.ComputeFaceNormals;
Var
  Lindex : integer;
  LVec1, LVec2 : TPoint3D;
begin
  try

    for Lindex := 0 to High(Faces) do with Faces[Lindex] do begin
      // find 2 vectors that lie on the plane
      LVec1 := Subtract(Vertices[Corners[1]].Point,
                        Vertices[Corners[2]].Point);
      LVec2 := Subtract(Vertices[Corners[1]].Point,
                        Vertices[Corners[0]].Point);

      Normal := UnitVector(CrossProduct(LVec1, LVec2));
    end;

  except
    ShowMessage('Exception in Model FaceNormal Method');
  end;
end;

// calculate vertex normals by averaging the face normals adjacent to them
procedure T3DModel.ComputeVertexNormals;
Var
  LVertex, LFace : integer;
begin
  try

    for LVertex := 0 to High(Vertices) do Vertices[LVertex].Normal := Origin;

    for LFace := 0 to High(Faces) do with Faces[LFace] do begin
      if Curved = False then Continue; // don't include flat faces
      for LVertex := 0 to High(Corners) do
        AddP(Vertices[Corners[LVertex]].Normal, Normal);
    end;

    // reduce to unit vectors
    for LVertex := 0 to High(Vertices) do with Vertices[LVertex] do 
      if (Normal.X <> 0) or (Normal.Y <> 0) or (Normal.Z <> 0) then
        UnitVectorP(Normal);

  except
    ShowMessage('Exception in Model ComputeVertexNormals Method');
  end;
end;

// create edges by stepping around faces
procedure T3DModel.AssignEdges;
Var
  LVertex, LEdge, LFace : integer;
  LTempEdge : TEdge;
  LEdgeDup : boolean;
begin
  try

    SetLength(Edges, 0);
    for LFace := 0 to High(Faces) do with Faces[LFace] do begin
      for LVertex := 0 to High(Corners) do begin
        LTempEdge.Point1 := Corners[LVertex];
        if LVertex = High(Corners) then LTempEdge.Point2 := Corners[0]
                                   else LTempEdge.Point2 := Corners[LVertex + 1];
        LEdgeDup := False;
        if length(Edges) > 0 then
          for LEdge := 0 to High(Edges) do
            if (Edges[LEdge].Point1 = LTempEdge.Point2) and
               (Edges[LEdge].Point2 = LTempEdge.Point1) then
              LEdgeDup := true;
        if not LEdgeDup then begin
          SetLength(Edges, length(Edges) + 1);
          Edges[High(Edges)] := LTempEdge;
        end;
      end;
    end;

  except
    ShowMessage('Exception in Model AssignEdges Method');
  end;
end;

// compute the center of all the faces in a model by averaging the vertices
procedure T3DModel.ComputeFaceCenters;
Var
  LFace, LVertex, LLen : integer;
  LSum : TPoint3D;
begin
  try

    for LFace := 0 to High(Faces) do with Faces[LFace] do begin
      LSum := Origin;
      LLen := length(Corners);
      for LVertex := 0 to LLen - 1 do
        AddP(LSum, Vertices[Corners[LVertex]].Point);
      Center := ScalarProduct(LSum, 1 / LLen);
    end;

  except
    ShowMessage('Exception in Model ComputeFaceCenters Method');
  end;
end;

// compute the bounding sphere of a model for frustum culling
procedure T3DModel.ComputeBoundingSphere;
Var
  LVertex : integer;
  LSum, LOffset : TPoint3D;
  LMaxR, LRadius : Real;
begin
  try

    // average the vertices coordinates to find bounding sphere center
    LSum := Origin;
    for LVertex := 0 to High(Vertices) do AddP(LSum, Vertices[LVertex].Point);
    BoundingSphere.Center := ScalarProduct(LSum, 1/length(Vertices));

    // find the vertex furthest from the center by calculating square of distance
    LMaxR := 0;
    for LVertex := 0 to High(Vertices) do begin
      LOffset := Subtract(Vertices[LVertex].Point, BoundingSphere.Center);
      LRadius := DotProduct(LOffset, LOffset);
      if LRadius > LMaxR then LMaxR := LRadius;
    end;

    // bounding sphere radius is then the square root of the max square distance
    BoundingSphere.Radius := Sqrt(LMaxR);

  except
    ShowMessage('Exception in Model ComputeBoundingSphere Method');
  end;
end;

// save the model data to a memory stream
procedure T3DModel.SaveToStream(AStream : TMemoryStream);
Var
  Lindex, LLen, LLenC : integer;
  SizeOfInteger : integer;
begin
  try

    SizeOfInteger := SizeOf(integer);

    with AStream do begin

      // write vertices
      LLen := length(Vertices);
      Write(LLen, SizeOfInteger);
      if LLen > 0 then for Lindex := 0 to LLen - 1 do
        Write(Vertices[Lindex].Point, SizeOf(TPoint3D));

      // write faces
      LLen := length(Faces);
      Write(LLen, SizeOfInteger);
      if LLen > 0 then for Lindex := 0 to LLen - 1 do begin
        LLenC := length(Faces[Lindex].Corners);
        Write(LLenC, SizeOfInteger);
        if LLenC > 0 then Write(Faces[Lindex].Corners[0], LLenC*SizeOfInteger);
        Write(Faces[Lindex].HLS, Sizeof(THLS));
        Write(Faces[LIndex].Curved, SizeOf(Boolean));
      end;

      // write edges
      LLen := length(Edges);
      Write(LLen, SizeOfInteger);
      if LLen > 0 then Write(Edges[0], LLen*SizeOf(TEdge));

      // write center position
      Write(Center, SizeOf(TPoint3D));

      // write color
      Write(Color, SizeOf(TRGB));

    end;

  except
    ShowMessage('Exception in Model SaveToStream Method');
  end;
end;

// load the model data from a memory stream
procedure T3DModel.LoadFromStream(AStream : TMemoryStream);
Var
  Lindex, LLen, LLenC : integer;
  SizeOfInteger : integer;
begin
  try

    SizeOfInteger := SizeOf(integer);

    with AStream do begin

      // read vertices
      Read(LLen, SizeOfInteger);
      SetLength(Vertices, LLen);
      if LLen > 0 then for Lindex := 0 to LLen - 1 do
        Read(Vertices[Lindex].Point, SizeOf(TPoint3D));

      // read faces
      Read(LLen, SizeOfInteger);
      SetLength(Faces, LLen);
      if LLen > 0 then for Lindex := 0 to LLen - 1 do begin
        Read(LLenC, SizeOfInteger);
        SetLength(Faces[Lindex].Corners, LLenC);
        if LLenC > 0 then Read(Faces[Lindex].Corners[0], LLenC*SizeOfInteger);
        Read(Faces[Lindex].HLS, Sizeof(THLS));
        Read(Faces[Lindex].Curved, SizeOf(Boolean));
      end;

      // read edges
      Read(LLen, SizeOfInteger);
      SetLength(Edges, LLen);
      if LLen > 0 then Read(Edges[0], LLen*SizeOf(TEdge));

      // read center position
      Read(Center, SizeOf(TPoint3D));

      // read color
      Read(Color, SizeOf(TRGB));

      // compute additional data
      ComputeFaceNormals;
      ComputeVertexNormals;
      ComputeFaceCenters;
      ComputeBoundingSphere;

    end;

  except
    ShowMessage('Exception in Model LoadFromStream Method');
  end;
end;

// save the model data to a file
procedure T3DModel.SaveToFile(APath : String);
Var
  LStream, LStreamPacked : TMemoryStream;
begin
  LStream := TMemoryStream.Create;
  LStream.Seek(0, 0);
  SaveToStream(LStream);
  LStreamPacked := TMemoryStream.Create;
  PackStream(LStream, LStreamPacked, clMax);
  LStream.Free;
  LStreamPacked.SaveToFile(APath);
  LStreamPacked.Free;
end;

// load the model data from a file
procedure T3DModel.LoadFromFile(APath : String);
Var
  LStream, LStreamPacked : TMemoryStream;
begin
  LStreamPacked := TMemoryStream.Create;
  LStreamPacked.LoadFromFile(APath);
  LStream := TMemoryStream.Create;
  ExpandStream(LStreamPacked, LStream);
  LStreamPacked.Free;
  LStream.Seek(0, 0);
  LoadFromStream(LStream);
  LStream.Free;
end;

end.
