(****************************************************************)
(*                                                              *)
(*             3D Rendering Engine Model Builder                *)
(*                                                              *)
(*                   Written by Peter Bone                      *)
(*                   peterbone@hotmail.com                      *)
(*           http://www.geocities.com/peter_bone_uk             *)
(*                                                              *)
(*                Last Modified  5 / 7 / 2005                   *)
(*                                                              *)
(*     This unit is a library of model building procedures      *)
(*          to create models for the rendering engine.          *)
(*                                                              *)
(*        This code is free for use but please give credit      *)
(*                                                              *)
(****************************************************************)

unit ModelBuild;

interface

uses
  math, dialogs, graphics, Model3D, ColorConv, Global;

type
  TVertexPair = record
    Vertex1 : integer;
    Vertex2 : integer;
    MidVertex : integer;
  end;

  TFloatArray = array of Real;
  T2DFloatArray = array of TFloatArray;

  TPlatonicSolid = (tetrahedron, octahedron, hexahedron, icosahedron, dodecahedron);

procedure BuildSolidOfRevolution(var AModel : T3DModel ;
                                 Coords, Radius : TFloatArray ;
                                 LayerColors : TColorArray ;
                                 Detail : integer);
procedure BuildSurface(var AModel : T3DModel ;
                       AHeights : T2DFloatArray ;
                       AXScale, AYSCale, AZScale : Real ;
                       AColor : TColor);
procedure BuildBox(var AModel : T3DModel ;
                   AXDim, AYDim, AZDim : integer ;
                   FaceColors : array of TColor);
procedure BuildPlatonicSolid(var AModel : T3DModel ;
                             APlatonicSolid : TPlatonicSolid ;
                             AScale : Real ;
                             AColor : TColor ;
                             CreateWireFrame : boolean);
procedure BuildGeoSphere(var AModel : T3DModel ;
                         AStartPlatonic : TPlatonicSolid ;
                         ADivisionLevel : byte ;
                         ARadius : Real;
                         AColor : TColor ;
                         Curved : Boolean;
                         CreateWireFrame : Boolean);
procedure BuildTube(var AModel : T3DModel ; Alength, ARadius : Real ;
                    ADivisionLevel : Byte ; AColor : TColor);

implementation

// build a general rotational object
procedure BuildSolidOfRevolution(var AModel : T3DModel ;
                                 Coords, Radius : TFloatArray ;
                                 LayerColors : TColorArray ;
                                 Detail : integer);
Var
  LAngle : Real;                    //              ___
  LRot : integer;                   //         _--''   ''--_
  LP : TPoints;                     //        /             \
  LFace, LEdge, LVertex : integer;  //       /`'--,,___,,--'/\
  Lsin, Lcos : Extended;            //      /`          ``.-//\
  LRing : integer;                  //      |`          ``.-//|
  NumRings : integer;               //       \`        ``.-/J/
begin                               //        \`       `.-/M/
try                                 //         \`     `.-/#/
                                    //          \`   `.-/#/
  with AModel do begin              //           |'-=///#|
                                    //          /`   `.-//\
    NumRings := length(Coords);     //         '--,,___,,##'
    SetLength(LP, NumRings);
    SetLength(Vertices, Detail * NumRings);
    SetLength(Edges, 2 * Detail * NumRings - Detail);
    SetLength(Faces, Detail * (NumRings - 1) + 2);
    LVertex := 0;
    LEdge := 0;
    LFace := 0;

    // set initial point positions and ring heights
    for LRing := 0 to NumRings - 1 do begin
      Vertices[LVertex].Point := Point3D(Radius[LRing], Coords[LRing], 0);
      Inc(LVertex);
      LP[LRing].Y := Coords[LRing];
    end;

    // create vertices by rotating points around rings
    for LRot := 1 to Detail do begin
      if LRot < Detail then begin
        LAngle := 2*pi * LRot / Detail;
        SinCos(LAngle, Lsin, Lcos);
        // calculate new point positions
        for LRing := 0 to NumRings - 1 do begin
          LP[LRing].X := Radius[LRing] * Lcos;
          LP[LRing].Z := Radius[LRing] * Lsin;
        end;
        // store new points in vertices
        for LRing := 0 to NumRings - 1 do begin
          Vertices[LVertex].Point := LP[LRing];
          Inc(LVertex);
        end;
      end;
      // create a new face and edges by indexing vertices
      for LRing := 0 to NumRings - 2 do begin
        SetLength(Faces[LFace].Corners, 4);
        Faces[LFace].Corners[0] := ((LRot-1) * NumRings) + LRing;
        Faces[LFace].Corners[1] := ((LRot-1) * NumRings) + LRing + 1;
        if LRot = Detail then begin // join up 1st layer and last layer faces
          Faces[LFace].Corners[2] := Faces[LRing].Corners[1];
          Faces[LFace].Corners[3] := Faces[LRing].Corners[0];
        end else begin
          Faces[LFace].Corners[2] := (LRot * NumRings) + LRing + 1;
          Faces[LFace].Corners[3] := (LRot * NumRings) + LRing;
        end;

        // create edges
        Edges[LEdge].Point1 := Faces[LFace].Corners[0];
        Edges[LEdge].Point2 := Faces[LFace].Corners[1];
        Edges[LEdge + 1].Point1 := Edges[LEdge].Point1;
        Edges[LEdge + 1].Point2 := Faces[LFace].Corners[3];
        Inc(LEdge, 2);

        // face colours
        Faces[LFace].HLS := RGBtoHLS( ColorToRGB(LayerColors[LRing]) );

        Faces[LFace].Curved := True;

        Inc(LFace);
      end;
    end;

    // create bottom edges
    LVertex := NumRings - 1;
    for LFace := 2 to Detail do begin
      Edges[LEdge].Point1 := LVertex;
      Inc(LVertex, NumRings);
      Edges[LEdge].Point2 := LVertex;
      Inc(LEdge);
    end;
    Edges[LEdge].Point1 := LVertex;
    Edges[LEdge].Point2 := NumRings - 1;

    // create end faces
    SetLength(Faces[High(Faces)-1].Corners, Detail);
    SetLength(Faces[High(Faces)].Corners, Detail);
    for LRot := 0 to Detail - 1 do begin
      Faces[High(Faces)-1].Corners[LRot] := LRot * NumRings;
      Faces[High(Faces)].Corners[Detail-1-LRot] := ((LRot+1) * NumRings) - 1;
    end;
    // colours
    Faces[High(Faces)-1].HLS := RGBtoHLS( ColorToRGB(LayerColors[High(LayerColors) - 1]) );
    Faces[High(Faces)].HLS := RGBtoHLS( ColorToRGB(LayerColors[High(LayerColors)]) );

    Faces[High(Faces)-1].Curved := False; // end faces are not curved
    Faces[High(Faces)].Curved := False;

    // calculate face normals
    ComputeFaceNormals;

    ComputeVertexNormals;

    ComputeFaceCenters;

    ComputeBoundingSphere;

  end; // with AModel

except
  ShowMessage('Exception in BuildSolidOfRevolution Method');
end;
end;

// build a 3D surface from a 2D array of grid heights
procedure BuildSurface(var AModel : T3DModel ;
                       AHeights : T2DFloatArray ;
                       AXScale, AYSCale, AZScale : Real ;
                       AColor : TColor);
Var
  LXIndex, LYIndex, LGridXLen, LGridYLen : integer;
  Lx, Ly : real;
  LVertex, LEdge, LFace : integer;
  LHLS : THLS;                //         ____________
begin                         //        /_/_/_/_/_/_/\ _
try                           //       /_/_/_/_/_/_/\/_/\
                              //      /_/_/_/_/_/_/\/_/\/\
  with AModel do begin        //     /_/_/_/_/_/_/\/_/\/\/
                              //                 \/_/\/\/
    LHLS := RGBtoHLS( ColorToRGB(AColor) );  //     \/\/
                                             //      \/
    // grid dimensions
    LGridYLen := length(AHeights[0]);
    LGridXLen := length(AHeights);

    // set vertices, edges and faces dimensions
    SetLength(Vertices, LGridYLen * LGridXLen);
    SetLength(Edges, LGridYLen * (LGridXLen - 1) + LGridXLen * (LGridYLen - 1));
    SetLength(Faces, (LGridYLen - 1) * (LGridXLen - 1));

    LVertex := 0;
    LEdge := 0;
    LFace := 0;

    // loop through grid coordinates
    for LYIndex := 0 to LGridYLen - 1 do begin
      Ly := LYIndex * AYScale;
      for LXIndex := 0 to LGridXLen - 1 do begin
        Lx := LXIndex * AXScale;

        // set vertex coordinate
        Vertices[LVertex].Point := Point3D(Lx, Ly, AHeights[LXIndex, LYIndex] * AZScale);

        if (LYIndex < LGridYLen - 1) and (LXIndex < LGridXLen - 1) then begin

          // create a face attached to this vertex
          with Faces[LFace] do begin
            SetLength(Corners, 4);
            Corners[0] := LVertex;
            Corners[1] := LVertex + 1;
            Corners[2] := LVertex + LGridXLen + 1;
            Corners[3] := LVertex + LGridXLen;
            Curved := True; // whole surface is curved
            HLS := LHLS; // set face color
          end; // with faces

          // create edges coming off this vertex
          Edges[LEdge].Point1 := LVertex;
          Edges[LEdge].Point2 := Faces[LFace].Corners[1];
          Edges[LEdge + 1].Point1 := LVertex;
          Edges[LEdge + 1].Point2 := Faces[LFace].Corners[3];
          Inc(LEdge, 2);

          Inc(LFace);
        end;

        Inc(LVertex);

      end; // x loop
    end; // y loop

    // create right and bottom lines of edges
    LVertex := (LGridYLen - 1) * LGridXLen;
    for LFace := 2 to LGridXLen do begin
      Edges[LEdge].Point1 := LVertex;
      Inc(LVertex);
      Edges[LEdge].Point2 := LVertex;
      Inc(LEdge);
    end;
    LVertex := LGridXLen - 1;
    for LFace := 2 to LGridYLen do begin
      Edges[LEdge].Point1 := LVertex;
      Inc(LVertex, LGridXLen);
      Edges[LEdge].Point2 := LVertex;
      Inc(LEdge);
    end;

    // calculate face normals
    ComputeFaceNormals;

    ComputeVertexNormals;

    ComputeFaceCenters;

    ComputeBoundingSphere;

  end; // with AModel

except
  ShowMessage('Exception in BuildSurface Method');
end;
end;

// build a box with specified dimensions and face colours
procedure BuildBox(var AModel : T3DModel ;
                   AXDim, AYDim, AZDim : integer ;
                   FaceColors : array of TColor);
Var
  LRot, LLevel, LVertex, LEdge, LFace : byte;
  LxInc, Lx, Ly, Lz : real;  
begin                          
try                            //    Face colour assignments
                               //                ___
  with AModel do begin         //               |4  |
                               //        ___ ___|___|___
    SetLength(Vertices, 8);    //       |0  |1  |2  |3  |
    SetLength(Edges, 12);      //       |___|___|___|___|
    SetLength(Faces, 6);       //               |5  |
    LVertex := 0;              //               |___|
    LEdge := 0;                
                               
    for LFace := 0 to 5 do SetLength(Faces[LFace].Corners, 4);

    Lx := 0;
    for LRot := 0 to 3 do begin
      LxInc := 0;
      if LRot = 0 then LxInc := AXDim;
      if LRot = 2 then LxInc := -AXDim;
      for LLevel := 0 to 1 do begin
        if LRot < 2 then Lz := 0 else Lz := AZDim;
        if LLevel = 0 then Ly := 0 else Ly := AYDim;
        Vertices[LVertex].Point := Point3D(Lx, Ly, Lz);
        Inc(LVertex);
      end;

      Lx := Lx + LxInc;

      Edges[LEdge].Point1 := LVertex - 2;
      Edges[LEdge].Point2 := LVertex - 1;
      Edges[LEdge + 1].Point1 := Edges[LEdge].Point1;
      Edges[LEdge + 2].Point1 := Edges[LEdge].Point2;
      Faces[LRot].Corners[0] := Edges[LEdge].Point1;
      Faces[LRot].Corners[1] := Edges[LEdge].Point2;
      Faces[4].Corners[LRot] := Edges[LEdge].Point1;
      Faces[5].Corners[3 - LRot] := Edges[LEdge].Point2;
      if LRot = 3 then begin
        Edges[LEdge + 1].Point2 := 0;
        Edges[LEdge + 2].Point2 := 1;
        Faces[LRot].Corners[2] := 1;
        Faces[LRot].Corners[3] := 0;
      end else begin
        Edges[LEdge + 1].Point2 := LVertex;
        Edges[LEdge + 2].Point2 := LVertex + 1;
        Faces[LRot].Corners[2] := Edges[LEdge + 2].Point2;
        Faces[LRot].Corners[3] := Edges[LEdge + 1].Point2;
      end;
      Inc(LEdge, 3);
    end;

    // other face parameters
    for LFace := 0 to 5 do begin
      Faces[LFace].Curved := False;
      Faces[LFace].HLS := RGBtoHLS( ColorToRGB(FaceColors[LFace]) );
    end;
    ComputeFaceNormals;

    BoundingSphere.Center := Point3D(AXDim * 0.5, AYDim * 0.5, AZDim * 0.5);
    BoundingSphere.Radius := Modulus(BoundingSphere.Center);

    ComputeFaceCenters;

  end; // with AModel

except
  ShowMessage('Exception in BuildBox Method');
end;
end;

procedure BuildPlatonicSolid(var AModel : T3DModel ;
                             APlatonicSolid : TPlatonicSolid ;
                             AScale : Real ;
                             AColor : TColor ;
                             CreateWireframe : boolean);
Const
  phi = 1.61803; // ...39887498948482                _----------_,
Var                                    //          ,"__         _-:,
  a, b, c : Real;                      //         /    ""--_--""...:\
  Lx, Ly, Lz, LVertex : byte;          //        /         |.........\
  LHLS : THLS;                         //       /          |..........\
  LPointArray : array[0..2] of real;   //      /,         _'_........./:
begin                                  //      ! -,    _-"   "-_... ,;;:
try                                    //      \   -_-"         "-_/;;;;
                                       //       \   \             /;;;;'
  with AModel do begin                 //        \   \           /;;;;
                                       //         '.  \         /;;;'
    if APlatonicSolid = tetrahedron then begin //   "-_\_______/;;'

      SetLength(Vertices, 4);                  //     dodecahedron
      SetLength(Faces, 4);
      for Lx := 0 to High(Faces) do SetLength(Faces[Lx].Corners, 3);
      Vertices[0].Point := Point3D( 0.5, 0.5, 0.5);
      Vertices[1].Point := Point3D(-0.5, 0.5,-0.5);
      Vertices[2].Point := Point3D(-0.5,-0.5, 0.5);
      Vertices[3].Point := Point3D( 0.5,-0.5,-0.5);
      Faces[0].Corners[0] := 3; Faces[0].Corners[1] := 1; Faces[0].Corners[2] := 0;
      Faces[1].Corners[0] := 3; Faces[1].Corners[1] := 2; Faces[1].Corners[2] := 1;
      Faces[2].Corners[0] := 2; Faces[2].Corners[1] := 3; Faces[2].Corners[2] := 0;
      Faces[3].Corners[0] := 1; Faces[3].Corners[1] := 2; Faces[3].Corners[2] := 0;

    end else
    if APlatonicSolid = octahedron then begin

      SetLength(Vertices, 6);
      SetLength(Faces, 8);
      for Lx := 0 to High(Faces) do SetLength(Faces[Lx].Corners, 3);
      a := 1 / (2 * sqrt(2));
      b := 0.5;
      Vertices[0].Point := Point3D(-a, 0, a);
      Vertices[1].Point := Point3D(-a, 0,-a);
      Vertices[2].Point := Point3D( a, 0,-a);
      Vertices[3].Point := Point3D( a, 0, a);
      Vertices[4].Point := Point3D( 0, b, 0);
      Vertices[5].Point := Point3D( 0,-b, 0);
      Faces[0].Corners[0] := 4; Faces[0].Corners[1] := 1; Faces[0].Corners[2] := 0;
      Faces[1].Corners[0] := 4; Faces[1].Corners[1] := 2; Faces[1].Corners[2] := 1;
      Faces[2].Corners[0] := 4; Faces[2].Corners[1] := 3; Faces[2].Corners[2] := 2;
      Faces[3].Corners[0] := 4; Faces[3].Corners[1] := 0; Faces[3].Corners[2] := 3;
      Faces[4].Corners[0] := 5; Faces[4].Corners[1] := 1; Faces[4].Corners[2] := 2;
      Faces[5].Corners[0] := 5; Faces[5].Corners[1] := 0; Faces[5].Corners[2] := 1;
      Faces[6].Corners[0] := 5; Faces[6].Corners[1] := 2; Faces[6].Corners[2] := 3;
      Faces[7].Corners[0] := 5; Faces[7].Corners[1] := 3; Faces[7].Corners[2] := 0;

    end else
    if APlatonicSolid = hexahedron then begin

      SetLength(Vertices, 8);
      SetLength(Faces, 6);
      for Lx := 0 to High(Faces) do SetLength(Faces[Lx].Corners, 4);
      LVertex := 0;
      for Lx := 0 to 1 do
        for Ly := 0 to 1 do
          for Lz := 0 to 1 do begin
            Vertices[LVertex].Point := Point3D(Lx-0.5, Ly-0.5, Lz-0.5);
            Inc(LVertex);
          end;
      Faces[0].Corners[0] := 0; Faces[0].Corners[1] := 4; Faces[0].Corners[2] := 5; Faces[0].Corners[3] := 1;
      Faces[1].Corners[0] := 0; Faces[1].Corners[1] := 1; Faces[1].Corners[2] := 3; Faces[1].Corners[3] := 2;
      Faces[2].Corners[0] := 1; Faces[2].Corners[1] := 5; Faces[2].Corners[2] := 7; Faces[2].Corners[3] := 3;
      Faces[3].Corners[0] := 2; Faces[3].Corners[1] := 3; Faces[3].Corners[2] := 7; Faces[3].Corners[3] := 6;
      Faces[4].Corners[0] := 4; Faces[4].Corners[1] := 6; Faces[4].Corners[2] := 7; Faces[4].Corners[3] := 5;
      Faces[5].Corners[0] := 0; Faces[5].Corners[1] := 2; Faces[5].Corners[2] := 6; Faces[5].Corners[3] := 4;

    end else
    if APlatonicSolid = icosahedron then begin

      SetLength(Vertices, 12);
      SetLength(Faces, 20);
      for Lx := 0 to High(Faces) do SetLength(Faces[Lx].Corners, 3);
      a := 0.5;
      b := 1 / (2 * phi);
      LVertex := 0;
      LPointArray[0] := 0;
      for Lx := 0 to 1 do
        for Ly := 0 to 1 do begin
          LPointArray[1] := (Lx*2-1)*b; LPointArray[2] := (Ly*2-1)*a;
          for Lz := 0 to 2 do begin
            Vertices[LVertex].Point := Point3D(LPointArray[(0+Lz) mod 3],
                                               LPointArray[(1+Lz) mod 3],
                                               LPointArray[(2+Lz) mod 3]);
            Inc(LVertex);
          end;
        end;
      Faces[0].Corners[0]  := 4;  Faces[0].Corners[1]  := 10; Faces[0].Corners[2]  :=6;
      Faces[1].Corners[0]  := 10; Faces[1].Corners[1]  := 4;  Faces[1].Corners[2]  :=9;
      Faces[2].Corners[0]  := 8;  Faces[2].Corners[1]  := 3;  Faces[2].Corners[2]  :=9;
      Faces[3].Corners[0]  := 3;  Faces[3].Corners[1]  := 11; Faces[3].Corners[2]  :=9;
      Faces[4].Corners[0]  := 5;  Faces[4].Corners[1]  := 0;  Faces[4].Corners[2]  :=6;
      Faces[5].Corners[0]  := 0;  Faces[5].Corners[1]  := 2;  Faces[5].Corners[2]  :=6;
      Faces[6].Corners[0]  := 1;  Faces[6].Corners[1]  := 7;  Faces[6].Corners[2]  :=3;
      Faces[7].Corners[0]  := 7;  Faces[7].Corners[1]  := 1;  Faces[7].Corners[2]  :=0;
      Faces[8].Corners[0]  := 2;  Faces[8].Corners[1]  := 8;  Faces[8].Corners[2]  :=4;
      Faces[9].Corners[0]  := 8;  Faces[9].Corners[1]  := 2;  Faces[9].Corners[2]  :=1;
      Faces[10].Corners[0] := 11; Faces[10].Corners[1] := 5;  Faces[10].Corners[2] :=10;
      Faces[11].Corners[0] := 5;  Faces[11].Corners[1] := 11; Faces[11].Corners[2] :=7;
      Faces[12].Corners[0] := 4;  Faces[12].Corners[1] := 8;  Faces[12].Corners[2] :=9;
      Faces[13].Corners[0] := 11; Faces[13].Corners[1] := 10; Faces[13].Corners[2] :=9;
      Faces[14].Corners[0] := 2;  Faces[14].Corners[1] := 4;  Faces[14].Corners[2] :=6;
      Faces[15].Corners[0] := 10; Faces[15].Corners[1] := 5;  Faces[15].Corners[2] :=6;
      Faces[16].Corners[0] := 1;  Faces[16].Corners[1] := 2;  Faces[16].Corners[2] :=0;
      Faces[17].Corners[0] := 5;  Faces[17].Corners[1] := 7;  Faces[17].Corners[2] :=0;
      Faces[18].Corners[0] := 8;  Faces[18].Corners[1] := 1;  Faces[18].Corners[2] :=3;
      Faces[19].Corners[0] := 7;  Faces[19].Corners[1] := 11; Faces[19].Corners[2] :=3;

    end else begin // dodecahedron

      SetLength(Vertices, 20);
      SetLength(Faces, 12);
      for Lx := 0 to High(Faces) do SetLength(Faces[Lx].Corners, 5);
      b := 1 / (2 * phi);
      c := (2 - phi) * 0.5;
      LVertex := 0;
      for Lx := 0 to 1 do
        for Ly := 0 to 1 do
          for Lz := 0 to 1 do begin
            Vertices[LVertex].Point := Point3D((2*Lx-1)*b, (2*Ly-1)*b, (2*Lz-1)*b);
            Inc(LVertex);
          end;
      LPointArray[0] := 0;
      for Lx := 0 to 1 do
        for Ly := 0 to 1 do begin
          LPointArray[1] := (Lx*2-1)*0.5; LPointArray[2] := (Ly*2-1)*c;
          for Lz := 0 to 2 do begin
            Vertices[LVertex].Point := Point3D(LPointArray[(0+Lz) mod 3],
                                               LPointArray[(1+Lz) mod 3],
                                               LPointArray[(2+Lz) mod 3]);
            Inc(LVertex);
          end;
        end;
      Faces[0].Corners[0]  := 7;  Faces[0].Corners[1]  := 17; Faces[0].Corners[2]  := 3; Faces[0].Corners[3]  := 16; Faces[0].Corners[4] := 19;
      Faces[1].Corners[0]  := 1;  Faces[1].Corners[1]  := 11; Faces[1].Corners[2]  := 5; Faces[1].Corners[3]  := 19; Faces[1].Corners[4] := 16;
      Faces[2].Corners[0]  := 4;  Faces[2].Corners[1]  := 8;  Faces[2].Corners[2]  := 0; Faces[2].Corners[3]  := 10; Faces[2].Corners[4] := 13;
      Faces[3].Corners[0]  := 2;  Faces[3].Corners[1]  := 14; Faces[3].Corners[2]  := 6; Faces[3].Corners[3]  := 13; Faces[3].Corners[4] := 10;
      Faces[4].Corners[0]  := 14; Faces[4].Corners[1]  := 17; Faces[4].Corners[2]  := 7; Faces[4].Corners[3]  := 18; Faces[4].Corners[4] := 6;
      Faces[5].Corners[0]  := 17; Faces[5].Corners[1]  := 14; Faces[5].Corners[2]  := 2; Faces[5].Corners[3]  := 12; Faces[5].Corners[4] := 3;
      Faces[6].Corners[0]  := 8;  Faces[6].Corners[1]  := 11; Faces[6].Corners[2]  := 1; Faces[6].Corners[3]  := 9;  Faces[6].Corners[4] := 0;
      Faces[7].Corners[0]  := 11; Faces[7].Corners[1]  := 8;  Faces[7].Corners[2]  := 4; Faces[7].Corners[3]  := 15; Faces[7].Corners[4] := 5;
      Faces[8].Corners[0]  := 7;  Faces[8].Corners[1]  := 19; Faces[8].Corners[2]  := 5; Faces[8].Corners[3]  := 15; Faces[8].Corners[4] := 18;
      Faces[9].Corners[0]  := 4;  Faces[9].Corners[1]  := 13; Faces[9].Corners[2]  := 6; Faces[9].Corners[3]  := 18; Faces[9].Corners[4] := 15;
      Faces[10].Corners[0] := 2;  Faces[10].Corners[1] := 10; Faces[10].Corners[2] := 0; Faces[10].Corners[3] := 9;  Faces[10].Corners[4] := 12;
      Faces[11].Corners[0] := 1;  Faces[11].Corners[1] := 16; Faces[11].Corners[2] := 3; Faces[11].Corners[3] := 12; Faces[11].Corners[4] := 9;

    end;

    // scale the model
    for LVertex := 0 to High(Vertices) do begin
      Vertices[LVertex].Point.X := Vertices[LVertex].Point.X * AScale;
      Vertices[LVertex].Point.Y := Vertices[LVertex].Point.Y * AScale;
      Vertices[LVertex].Point.Z := Vertices[LVertex].Point.Z * AScale;
    end;
                                                   //            .
    // face properties                             //           /!;,
    LHLS := RGBToHLS(ColorToRGB(AColor));          //          / `;;,
    for Lx := 0 to High(Faces) do begin            //         /   !;;;,
      Faces[Lx].Curved := False;                   //        /    `;;;;,
      Faces[Lx].HLS := LHLS;                       //       /      !;;;;,
    end;                                           //      /       `;;;;;)
    ComputeFaceNormals;                            //     /         !;;;/
                                                   //     ``--__    `:;/
    if CreateWireframe then AssignEdges;           //           ``--_!/

    BoundingSphere.Center := Origin;               //       tetrahedron
    BoundingSphere.Radius := Modulus(Vertices[0].Point);

    ComputeFaceCenters;

  end; // with AModel

except
  ShowMessage('Exception in BuildPlatonicSolid Method');
end;
end;

// build a geodesic sphere
procedure BuildGeoSphere(var AModel : T3DModel ;
                         AStartPlatonic : TPlatonicSolid ;
                         ADivisionLevel : byte ;
                         ARadius : Real; 
                         AColor : TColor ; 
                         Curved : Boolean;
                         CreateWireFrame : boolean);
Var
  LFace, LVertex : integer;
  LHLS : THLS;
  Lindex : integer;                     //        /\      each face of
  LLastFace : integer;                  //       /  \     a platonic
  LVertex1, LVertex2 : integer;         //      /____\    solid will
  LVertexPair : integer;                //     /\    /\   be divided
  LDupVertex : boolean;                 //    /  \  /  \  like this
  LMidVertex : array[0..2] of integer;  //   /____\/____\ 
  LLevel : byte;
  LVertexPairs : array of TVertexPair;
begin
  { number of faces for a given platonic and division level
             ________________________
   DIVISION | 0  | 1  | 2  | 3  | 4  |
            |____|____|____|____|____|
      tetra | 4  | 16 | 64 |256 |1024| face sizes vary a lot
            |____|____|____|____|____|
      octa  | 8  | 32 |128 |512 |2048| face sizes vary a bit
            |____|____|____|____|____|
      icosa | 20 | 80 |320 |1280|5120| face sizes almost the same
            |____|____|____|____|____|   }
try

  // check for valid platonic solid
  if (AStartPlatonic = hexahedron) or (AStartPlatonic = dodecahedron) then begin
    ShowMessage('AStartPlatonic must have triangular faces');
    Exit;
  end;

  with AModel do begin 

    // build the starting platonic solid 
    BuildPlatonicSolid(AModel, AStartPlatonic, 1, clwhite, False);   

    // scale to radius = 1
    for LVertex := 0 to High(Vertices) do 
      Vertices[LVertex].Point := UnitVector(Vertices[LVertex].Point);

    // loop through sub-division levels  
    if ADivisionLevel > 0 then for LLevel := 1 to ADivisionLevel do begin
    
      SetLength(LVertexPairs, 0);

      // sub-divide faces
      for LFace := 0 to High(Faces) do begin

        // create 3 new vertices at midpoint of triangle edges
        for Lindex := 0 to 2 do begin
          LDupVertex := False;
          LVertex1 := Faces[LFace].Corners[Lindex];
          LVertex2 := Faces[LFace].Corners[(Lindex + 1) mod 3];
          // check if this vertex already exists
          if length(LVertexPairs) > 0 then begin
            for LVertexPair := High(LVertexPairs) downto 0 do begin
              if (LVertex1 = LVertexPairs[LVertexPair].Vertex2) and
                 (LVertex2 = LVertexPairs[LVertexPair].Vertex1) then begin
                LDupVertex := True;
                LMidVertex[Lindex] := LVertexPairs[LVertexPair].MidVertex;
                // pairs can only be shared by 2 faces - so can now delete this one
                LVertexPairs[LVertexPair] := LVertexPairs[High(LVertexPairs)];
                SetLength(LVertexPairs, High(LVertexPairs));
                Break;
              end;
            end;
          end;
          if not LDupVertex then begin
            // doesn't exist so create and log in vertex pairs array
            SetLength(Vertices, length(Vertices) + 1);
            SetLength(LVertexPairs, length(LVertexPairs) + 1);
            LVertexPairs[High(LVertexPairs)].Vertex1 := LVertex1;
            LVertexPairs[High(LVertexPairs)].Vertex2 := LVertex2;
            LVertexPairs[High(LVertexPairs)].MidVertex := High(Vertices);
            LMidVertex[Lindex] := High(Vertices);
            Vertices[LMidVertex[Lindex]].Point := UnitVector(
              MidPoint(Vertices[LVertex1].Point, Vertices[LVertex2].Point) );
          end;
        end;

        // create 4 child faces from the parent face
        SetLength(Faces, length(Faces) + 3);
        LLastFace := High(Faces);
        for Lindex := LLastFace - 2 to LLastFace do
          SetLength(Faces[Lindex].Corners, 3);
        Faces[LLastFace - 2].Corners[0] := LMidVertex[0];
        Faces[LLastFace - 2].Corners[1] := Faces[LFace].Corners[1];
        Faces[LLastFace - 2].Corners[2] := LMidVertex[1];
        Faces[LLastFace - 1].Corners[0] := LMidVertex[2];
        Faces[LLastFace - 1].Corners[1] := LMidVertex[1];
        Faces[LLastFace - 1].Corners[2] := Faces[LFace].Corners[2];
        for Lindex := 0 to 2 do
          Faces[LLastFace].Corners[Lindex] := LMidVertex[Lindex];
        Faces[LFace].Corners[1] := LMidVertex[0];
        Faces[LFace].Corners[2] := LMidVertex[2];

      end; // sub-dividing this level of faces

    end; // looping through sub-division levels

    // scale the geo-sphere to the required radius
    for LVertex := 0 to High(Vertices) do
      ScalarProductP(Vertices[LVertex].Point, ARadius);

    // face parameters
    LHLS := RGBToHLS( ColorToRGB(AColor) );
    for LFace := 0 to High(Faces) do begin
      Faces[LFace].Curved := Curved;
      Faces[LFace].HLS := LHLS;
    end;
    ComputeFaceNormals;

    BoundingSphere.Center := Origin;
    BoundingSphere.Radius := ARadius;

    ComputeVertexNormals;

    if CreateWireFrame then AssignEdges; // find and store edges

    ComputeFaceCenters;

  end; // with AModel

except
  ShowMessage('Exception in BuildGeoSphere Method');
end;
end;

// build a tube with geo-sphere ends
procedure BuildTube(var AModel : T3DModel ; Alength, ARadius : Real ;
                    ADivisionLevel : Byte ; AColor : TColor);

  procedure QuickSortTubeEdges(var A: array of TPointFloat; iLo, iHi: Integer);
  Var
    Lo, Hi : Integer;
    Mid : Real;
    T : TPointFloat;
  begin
    Lo := iLo;
    Hi := iHi;
    Mid := A[(Lo + Hi) shr 1].Y;
    repeat
      while A[Lo].Y < Mid do Inc(Lo);
      while A[Hi].Y > Mid do Dec(Hi);
      if Lo <= Hi then begin
        T := A[Lo];
        A[Lo] := A[Hi];
        A[Hi] := T;
        Inc(Lo);
        Dec(Hi);
      end;
    until Lo > Hi;
    if Hi > iLo then QuickSortTubeEdges(A, iLo, Hi);
    if Lo < iHi then QuickSortTubeEdges(A, Lo, iHi);
  end;

Var
  LVertex, LFace, LCorner, HiF, HiV, LNewI, LLen : integer;
  Lend : boolean;
  LEdgeV : array of TPointFloat;
  LPoint : TPointFloat;
  LSum : TPoint3D;
begin
  with AModel do begin

    // build a geo-sphere
    BuildGeoSphere(AModel, octahedron, ADivisionLevel, ARadius, AColor, True, False);

    // create hemi-geo-sphere
    // remove lower faces
    LFace := 0;
    LEnd := False;
    while LFace <= High(Faces) do begin
      while not LEnd and (Faces[LFace].Center.Y > 0) do begin // bottom half
        HiF := High(Faces);
        Faces[LFace] := Faces[HiF];
        if LFace = HiF then LEnd := True;
        SetLength(Faces, HiF);
      end;
      Inc(LFace);
    end;

    // remove lower vertices
    LVertex := 0;
    LEnd := False;
    while LVertex <= High(Vertices) do begin
      while not LEnd and (Vertices[LVertex].Point.Y > 0) do begin
        HiV := High(Vertices);
        Vertices[LVertex] := Vertices[HiV];
        // correct corner indexing to the moved vertex
        for LFace := 0 to High(Faces) do begin
          for LCorner := 0 to High(Faces[LFace].Corners) do
            if Faces[LFace].Corners[LCorner] = HiV then
              Faces[LFace].Corners[LCorner] := LVertex;
        end;
        if LVertex = HiV then LEnd := True;
        SetLength(Vertices, HiV);
      end;
      Inc(LVertex);
    end;

    HiV := High(Vertices);

    // find edge vertices and sort in order of angle around center
    SetLength(LEdgeV, 0);
    for LVertex := 0 to HiV do begin
      if Vertices[LVertex].Point.Y = 0 then begin
        SetLength(LEdgeV, length(LEdgeV)+1);
        LEdgeV[High(LEdgeV)].X := LVertex;
        LPoint.X := Vertices[LVertex].Point.X;
        LPoint.Y := Vertices[LVertex].Point.Z;
        LEdgeV[High(LEdgeV)].Y := CalcAngle(LPoint, Point2D(0, 0));
      end;
    end;
    QuickSortTubeEdges(LEdgeV, 0, High(LEdgeV));
    LLen := length(LEdgeV);

//    for LVertex := 0 to LLEn-1 do
//      ScalarProductP(Vertices[Round(LEdgeV[LVertex].X)].Normal, LVertex/3);

    // duplicate hemisphere and turn inside out
    // duplicate faces, invert and shift by the tube length
    HiF := High(Faces);
    SetLength(Faces, 2*(HiF+1));
    for LFace := 0 to HiF do begin
      LNewI := HiF+LFace+1;
      AssignFace(Faces[LNewI], Faces[LFace]);
      Faces[LNewI].Normal.Y := -Faces[LNewI].Normal.Y;
      Faces[LNewI].Center.Y := ALength-Faces[LNewI].Center.Y;
      // correct vertex indexing of duplicate faces
      for LCorner := 0 to High(Faces[LNewI].Corners) do
        Inc(Faces[LNewI].Corners[LCorner], HiV+1);
    end;

    // duplicate vertices, invert and shift by the tube length
    SetLength(Vertices, 2*(HiV+1));
    for LVertex := 0 to HiV do begin
      LNewI := HiV+LVertex+1;
      Vertices[LNewI] := Vertices[LVertex];
      Vertices[LNewI].Normal.Y := -Vertices[LNewI].Normal.Y;
      Vertices[LNewI].Point.Y := ALength-Vertices[LNewI].Point.Y;
    end;

    // create faces to join hemispheres
    HiF := High(Faces);
    SetLength(Faces, HiF+1+LLen);
    for LFace := HiF+1 to High(Faces) do begin
      LNewI := LFace - HiF - 1;
      Faces[LFace].Curved := Faces[0].Curved;
      Faces[LFace].HLS := Faces[0].HLS;
      SetLength(Faces[LFace].Corners, 4);
      Faces[LFace].Corners[3] := Round(LEdgeV[LNewI].X);
      Faces[LFace].Corners[2] := Round(LEdgeV[(LNewI+1) mod LLen].X);
      Faces[LFace].Corners[1] := Round(LEdgeV[(LNewI+1) mod LLen].X) + HiV + 1;
      Faces[LFace].Corners[0] := Round(LEdgeV[LNewI].X) + HiV + 1;
      LSum := Origin;
      for LCorner := 0 to 3 do
        AddP(LSum, Vertices[Faces[LFace].Corners[LCorner]].Point);
      Faces[LFace].Center := ScalarProduct(LSum, 0.25);
    end;
    ComputeFaceNormals;

    BoundingSphere.Center := Point3D(0, ALength*0.5, 0);
    BoundingSphere.Radius := ALength*0.5 + ARadius;

    AssignEdges;
  end;

  
end;

end.
