(*****************************************************************)
(*                                                               *)
(*              3D Rendering Engine World class                  *)
(*                                                               *)
(*                   Written by Peter Bone                       *)
(*                   peterbone@hotmail.com                       *)
(*           http://www.geocities.com/peter_bone_uk              *)
(*                                                               *)
(*                Last Modified  5 / 7 / 2005                    *)
(*                                                               *)
(*    This unit contains the T3DWorld class used to store the    *)
(*    properties of the world and the objects in it and also     *)
(*                      render the world                         *)
(*                                                               *)
(*       This code is free for use but please give credit        *)
(*                                                               *)
(*****************************************************************)

unit World3D;

interface

uses
  Windows, Classes, Model3D, Dialogs, Graphics, ColorConv, Global, Camera,
  Math, ZLib;

type

  TRGBArray = array[0..32767] of TRGB;
  pRGBArray = ^TRGBArray;

  TScanlines = array of pRGBArray;

  TRGBFixed = record
    R : integer;
    G : integer;
    B : integer;
  end;

  TPointColor = record
    X : single;
    Y : single;
    RGB : TRGB;
  end;
  TPointColorTriangle = array[0..2] of TPointColor;

  TPointLight = record
    X : single;
    Y : single;
    N : TPoint3DSingle;
    S : TPoint3DSingle;
  end;
  TPointLightTriangle = array[0..2] of TPointLight;

  TPointTriangle = array[0..2] of TPointFloat;

  TLight = record
    Position : TPoint3D; // position of light in world
    Luminance : byte; // brightness
  end;

  TVisibleIndex = record
    ObjectIndex : integer;
    Index : integer;
    Distance : Real; // distance from camera - for sorting
  end;
  TVisibleIndexArray = array of TVisibleIndex;

  T3DWorld = Class(TObject)
    private

      FWireframe : boolean;
      FPhong : boolean;
      FBackfaceRemoval : boolean;
      FAmbient : byte; // ambient light intensity
      FScreenRect : TRect;
      FSuperRect : TRect;
      FBackgroundColor : TColor;
      FSuperSample : byte; // power of 2 to supersample by 0, 2, 4, 6, ...
      FSuperRatio, FSSshr1 : byte;
      ScanlinesOSBmp : TScanlines;
      Scanlines : TScanlines;
      SuperBmp : TBitmap;

      procedure SetScreenRect(ARect : TRect);
      procedure SetSuperSample(ASample : byte);
      procedure RenderWireframe;
      procedure SuperSampleBitmap;
      procedure QuickSort(var A: TVisibleIndexArray; iLo, iHi: Integer);
      function PerspectiveProject(APoint3D : TPoint3D) : TPointFloat;
      procedure FlatFace(AObject, AFace : integer);
      procedure GouraudFace(AObject, AFace : integer);
      procedure PhongFace(AObject, AFace : integer);
      procedure GouraudPoly(W, H : integer ; V : TPointColorTriangle);
      procedure PhongPoly(W, H : integer ; V : TPointLightTriangle ;
                          Hue, Sat, Lum, Amb : byte);
      procedure WuLine(x1, y1, x2, y2, W, H : integer ; R, G, B : byte);

    public

      Objects : array of T3DModel; // array of objects in world
      Camera : TCamera;
      Light : TLight;
      OffScrBmp : TBitmap; // off screen bitmap world will be rendered to

      procedure Assign(const Source : T3DWorld);
      procedure Render;
      procedure ClearBitmap;
      procedure SaveToStream(AStream : TMemoryStream);
      procedure LoadFromStream(AStream : TMemoryStream);
      procedure SaveToFile(APath : String);
      procedure LoadFromFile(APath : String);

      property ScreenRect : TRect read FScreenRect write SetScreenRect;
      property Wireframe : boolean read FWireframe write FWireframe;
      property Phong : boolean read FPhong write FPhong;
      property BackfaceRemoval : boolean read FBackFaceRemoval write FBackFaceRemoval;
      property Ambient : byte read FAmbient write FAmbient;
      property BackgroundColor : TColor read FBackgroundColor write FBackgroundColor;
      property SuperSample : byte read FSuperSample write SetSuperSample;

      constructor Create;
      destructor Destroy; override;

    end;

implementation

constructor T3DWorld.Create;
begin
  // set parameters to default
  BackfaceRemoval := True;
  Wireframe := False;
  Phong := False;
  BackgroundColor := clBlack;
  SetLength(Objects, 0);

  // create the off-screen bitmap for rendering
  OffScrBmp := TBitmap.Create;
  OffScrBmp.PixelFormat := pf32bit;

  // no super-sampling by default
  FSuperSample := 0;

  // create the camera
  Camera := TCamera.Create;

  Origin := Point3D(0, 0, 0);
end;

destructor T3DWorld.Destroy;
Var
  LObject : integer;
begin
  if length(Objects) > 0 then // free the objects
    for LObject := 0 to High(Objects) do Objects[LObject].Free;

  Camera.Free;
  OffScrBmp.Free;
  SuperBmp.Free;

  inherited Destroy;
end;

procedure T3DWorld.Assign(const Source : T3DWorld);
Var
  LObject : integer;
begin
  ScreenRect := Source.ScreenRect;
  Phong := Source.Phong;
  WireFrame := Source.Wireframe;
  BackgroundColor := Source.BackgroundColor;
  SuperSample := Source.SuperSample;
  BackfaceRemoval := Source.BackfaceRemoval;
  Ambient := Source.Ambient;
  Light := Source.Light;
  Camera.Assign(Source.Camera);

  // destroy objects
  if length(Objects) > length(Source.Objects) then
    for LObject := length(Source.Objects) to High(Objects) do
      Objects[LObject].Free;
  // assign new objects
  SetLength(Objects, length(Source.Objects));
  for LObject := 0 to High(Objects) do begin
    Objects[LObject] := T3DModel.Create;
    Objects[LObject].Assign(Source.Objects[LObject]);
  end;
end;

procedure T3DWorld.SetScreenRect(ARect : TRect);
begin
  FScreenRect := ARect;

  // set screen center position
  Camera.ScreenMid.X := ARect.Right  shr 1;
  Camera.ScreenMid.Y := ARect.Bottom shr 1;

  // set bitmap dimensions
  OffScrBmp.Width  := ARect.Right;
  OffScrBmp.Height := ARect.Bottom;

  // update super-sampling bitmap and scanlines
  SuperSample := FSuperSample;
end;

// set SuperSample after ScreenRect
procedure T3DWorld.SetSuperSample(ASample : byte);
Var
  LLine, LRowSize : integer;
  LRowP : pRGBArray;
begin
  if Odd(ASample) then FSuperSample := ASample - 1
                  else FSuperSample := ASample;
  FSSshr1 := ASample shr 1;
  FSuperRatio := 1 shl FSSshr1;
  if FSuperSample > 0 then begin
    // set bitmaps and scanlines for super-sampling
    FSuperRect.Left := 0; FSuperRect.Top := 0;
    FSuperRect.Right  := ScreenRect.Right  shl FSSshr1;
    FSuperRect.Bottom := ScreenRect.Bottom shl FSSshr1;
    SuperBmp := TBitmap.Create;
    SuperBmp.PixelFormat := pf32bit;
    SuperBmp.Width  := FSuperRect.Right;
    SuperBmp.Height := FSuperRect.Bottom;
    SetLength(ScanlinesOSBmp, FScreenRect.Bottom);
    LRowSize := -BytesPerScanline(FScreenRect.Right, 32, 32);
    LRowP := OffScrBmp.ScanLine[0];
    for LLine := 0 to FScreenRect.Bottom - 1 do begin
      ScanlinesOSBmp[LLine] := LRowP;
      Inc(Integer(LRowP), LRowSize);
    end;
    SetLength(Scanlines, FSuperRect.Bottom);
    LRowSize := -BytesPerScanline(FSuperRect.Right, 32, 32);
    LRowP := SuperBmp.ScanLine[0];
    for LLine := 0 to FSuperRect.Bottom - 1 do begin
      Scanlines[LLine] := LRowP;
      Inc(Integer(LRowP), LRowSize);
    end;
  end else begin
    // set bitmaps and scanlines for no super-sampling
    SetLength(Scanlines, FScreenRect.Bottom);
    LRowSize := -BytesPerScanline(FScreenRect.Right, 32, 32);
    LRowP := OffScrBmp.ScanLine[0];
    for LLine := 0 to FScreenRect.Bottom - 1 do begin
      Scanlines[LLine] := LRowP;
      Inc(Integer(LRowP), LRowSize);
    end;
    SuperBmp.Free;
  end;
end;

// transform the coordinates of the objects to view space and render each one
procedure T3DWorld.Render;
Var
  LObject, LFaceIndex, LCorner : integer;
  LVertex, LFace, LNumFaces : integer;
  LIntensityRatio : Real;
  LSourceDirection, LDistanceVector : TPoint3D;
  LHiCorners : integer;
  LFrustumState, LPlane : byte;
  LFaceInside : boolean;
  LVisibleFaces : TVisibleIndexArray;
//  LNormalTo : TPoint3D;
//  LNormalToScreen : TPointFloat;
begin
  try

    // calculate frustum for clipping
    Camera.PositionFrustum;

    // copy bitmap to the super-sampling bitmap
    if SuperSample > 0 then SuperBmp.Canvas.StretchDraw(FSuperRect, OffScrBmp);

    if Wireframe then begin
      RenderWireframe;
      if SuperSample > 0 then SuperSampleBitmap;
      Exit;
    end;

    // create a single face array for all objects that indexes an object and
    // face. This is to sort and render all faces together so that they intersect
    SetLength(LVisibleFaces, 0);
    LNumFaces := 0;

    for LObject := 0 to High(Objects) do with Objects[LObject] do begin

      LFrustumState := Camera.SphereInFrustum(BoundingSphere);
      // skip objects that are completely outside the field of vision (FOV)
      if LFrustumState = Outside then Continue;

      // copy faces
      for LFace := 0 to High(Faces) do with Faces[LFace] do begin

        // calculate distance of face from camera
        LDistanceVector := Subtract(Center, Camera.Position);
        
        // if face normal is pointing towards camera then it is visible else it is invisible
        if BackfaceRemoval and
           (DotProduct(Normal, UnitVector(LDistanceVector)) > 0) then Continue;

        LHiCorners := High(Corners);
        // check for face inside viewing frustum if object is intersecting
        if LFrustumState = Intersecting then begin
          for LPlane := 0 to 5 do begin
            LFaceInside := False;
            for LCorner := 0 to LHiCorners do
              if Camera.PointPlaneSide(LPlane, Vertices[Corners[LCorner]].Point) then
                LFaceInside := True;
            if not LFaceInside then Break;
          end;
          if not LFaceInside then Continue; // face is completely outside FOV
        end;

        // add face to visible faces array
        Inc(LNumFaces);
        SetLength(LVisibleFaces, LNumFaces);
        LVisibleFaces[LNumFaces-1].ObjectIndex := LObject;
        LVisibleFaces[LNumFaces-1].Index := LFace;

        // calculate face distance from camera
        // no need for slow Sqrt, so not Euclidean distance
        LVisibleFaces[LNumFaces-1].Distance := DotProduct(LDistanceVector,
                                                          LDistanceVector);

      end; // face loop

    end; // object loop

    if LNumFaces = 0 then Exit; // nothing is visible

    // calculate the luminance and screen coordinates of the visible vertices
    // this means that they will only have to be calculated once for each vertex
    for LObject := 0 to High(Objects) do
      for LVertex := 0 to High(Objects[LObject].Vertices) do
        Objects[LObject].Vertices[LVertex].Visible := False;
    for LFace := 0 to LNumFaces - 1 do begin
      with Objects[LVisibleFaces[LFace].ObjectIndex] do begin
        LFaceIndex := LVisibleFaces[LFace].Index;
        Faces[LFaceIndex].ViewCenter := Subtract(Faces[LFaceIndex].Center,
                                                 Camera.Position);
        Camera.LookAt(Faces[LFaceIndex].ViewCenter); // convert to view space
        for LCorner := 0 to High(Faces[LFaceIndex].Corners) do begin
          LVertex := Faces[LFaceIndex].Corners[LCorner];
          if not Vertices[LVertex].Visible then with Vertices[LVertex] do begin
            Visible := True; // flag it so it won't be recalculated
            if not Phong then begin
              LSourceDirection := UnitVector( Subtract(Light.Position, Point) );
              LIntensityRatio := DotProduct(Normal, LSourceDirection);
              if LIntensityRatio < 0 then Lum := Ambient else
                Lum := Min(Ambient + Round(Light.Luminance * LIntensityRatio), 255);
            end;
            ViewPoint := Subtract(Point, Camera.Position);
            Camera.LookAt(ViewPoint);
            ScreenPoint := PerspectiveProject(ViewPoint);
          end;
        end;
      end;
    end;

    // sort faces by distance
    QuickSort(LVisibleFaces, 0, LNumFaces - 1);

    // draw faces in order of furthest to closest so that nearer faces are drawn last
    for LFace := LNumFaces - 1 downto 0 do with LVisibleFaces[LFace] do begin
      if Objects[ObjectIndex].Faces[Index].Curved then begin
        if Phong then PhongFace(ObjectIndex, Index)
                 else GouraudFace(ObjectIndex, Index)
      end else
        FlatFace(ObjectIndex, Index);
    end;

    // super-sample for smoothing
    if SuperSample > 0 then SuperSampleBitmap;

    // debugging - draw vertex normals 
//    for LObject := 0 to High(Objects) do with Objects[LObject] do begin
//      OffScrBmp.Canvas.Pen.Color := clyellow;
//      for LVertex := 0 to High(Vertices) do begin
//        if not Vertices[LVertex].Visible then Continue;
//        LNormalTo := Add(Vertices[LVertex].Point, ScalarProduct(Vertices[LVertex].Normal, 15));
//        LNormalTo := Subtract(LNormalTo, Camera.Position);
//        Camera.LookAt(LNormalTo);
//        LNormalToScreen := PerspectiveProject(LNormalTo);
//        OffScrBmp.Canvas.MoveTo(round(Vertices[LVertex].ScreenPoint.X), round(Vertices[LVertex].ScreenPoint.Y));
//        OffScrBmp.Canvas.LineTo(round(LNormalToScreen.X), round(LNormalToScreen.Y));
//      end;
//    end;

  except
    ShowMessage('Exception in World Render Method');
  end;
end;

// render the world as wireframe - draw edges only
procedure T3DWorld.RenderWireframe;
Var
  LObject, LEdge, LVertex, LNumEdges : integer;
  LFrustumState, LPlane : byte;
  LEdgeInside : boolean;
  LVisibleEdges : TVisibleIndexArray;
  LDistanceVector : TPoint3D;
  P1, P2 : TPointFloat;
begin
  try

    SetLength(LVisibleEdges, 0);
    LNumEdges := 0;

    // loop through objects
    for LObject := 0 to High(Objects) do with Objects[LObject] do begin

      LFrustumState := Camera.SphereInFrustum(BoundingSphere);
      // skip objects that are completely outside the field of vision (FOV)
      if LFrustumState = Outside then Continue;

      // calculate vertex screen projections
      for LVertex := 0 to High(Vertices) do with Vertices[LVertex] do begin
        ViewPoint := Subtract(Point, Camera.Position);
        Camera.LookAt(ViewPoint);
        ScreenPoint := PerspectiveProject(ViewPoint);
      end;

      for LEdge := 0 to High(Edges) do with Edges[LEdge] do begin
        // check for edge inside viewing frustum if object is intersecting
        if LFrustumState = Intersecting then begin
          for LPlane := 0 to 5 do begin
            LEdgeInside := False;
            if Camera.PointPlaneSide(LPlane, Vertices[Point1].Point) then
              LEdgeInside := True;
            if Camera.PointPlaneSide(LPlane, Vertices[Point2].Point) then
              LEdgeInside := True;
            if not LEdgeInside then Break;
          end;
          if not LEdgeInside then Continue; // edge is completely outside FOV
        end;

        SetLength(LVisibleEdges, LNumEdges + 1);
        LVisibleEdges[LNumEdges].ObjectIndex := LObject;
        LVisibleEdges[LNumEdges].Index := LEdge;
        LDistanceVector := Subtract(MidPoint(Vertices[Point1].ViewPoint,
                                             Vertices[Point2].ViewPoint),
                                    Camera.Position);
        LVisibleEdges[LNumEdges].Distance := DotProduct(LDistanceVector,
                                                        LDistanceVector);
        Inc(LNumEdges);
      end; // edge loop

    end; // object loop

    if LNumEdges = 0 then Exit;

    // sort edges by distance
    QuickSort(LVisibleEdges, 0, LNumEdges - 1);

    // draw edges from back to front - anti-aliased
    for LEdge := LNumEdges - 1 downto 0 do with LVisibleEdges[LEdge] do begin
      with Objects[ObjectIndex] do begin
        P1 := Vertices[Edges[Index].Point1].ScreenPoint;
        P2 := Vertices[Edges[Index].Point2].ScreenPoint;
        WuLine(Round(P1.X) shl FSSshr1, Round(P1.Y) shl FSSshr1,
               Round(P2.X) shl FSSshr1, Round(P2.Y) shl FSSshr1,
               ScreenRect.Right shl FSSshr1, ScreenRect.Bottom shl FSSshr1,
               Color.R, Color.G, Color.B);
      end;
    end;

  except
    ShowMessage('Exception in World RenderWireframe Method');
  end;
end;

// sort faces / edges by Z for ordered rendering
procedure T3DWorld.QuickSort(var A: TVisibleIndexArray; iLo, iHi: Integer);
Var
  Lo, Hi : Integer;
  Mid : Real;
  T : TVisibleIndex;
begin
  Lo := iLo;
  Hi := iHi;
  Mid := A[(Lo + Hi) shr 1].Distance;
  repeat
    while A[Lo].Distance < Mid do Inc(Lo);
    while A[Hi].Distance > Mid do Dec(Hi);
    if Lo <= Hi then
    begin
      T := A[Lo];
      A[Lo] := A[Hi];
      A[Hi] := T;
      Inc(Lo);
      Dec(Hi);
    end;
  until Lo > Hi;
  if Hi > iLo then QuickSort(A, iLo, Hi);
  if Lo < iHi then QuickSort(A, Lo, iHi);
end;

// map a 3D point onto the 2D screen with correct perspective - uses similar triangles
function T3DWorld.PerspectiveProject(APoint3D : TPoint3D) : TPointFloat;
Var
  LCoordRatio : Real;
begin
  try

    LCoordRatio := Camera.PerspectiveOffset / APoint3D.Z;
    Result.X := (LCoordRatio * APoint3D.X) + Camera.ScreenMid.X;
    Result.Y := (LCoordRatio * APoint3D.Y) + Camera.ScreenMid.Y;

  except
    ShowMessage('Exception in World PerspectiveProject method');
  end;
end;

// render a face of an object with a given colour
procedure T3DWorld.FlatFace(AObject, AFace : integer);
Var
  LPolygon : array of TPoint;
  LColor : TColor;
  LIntensityRatio : Real;
  LVertex : byte;
  LSourceDirection : TPoint3D;
begin
  try

    with Objects[AObject] do begin

      SetLength(LPolygon, length(Faces[AFace].Corners));

      LSourceDirection := UnitVector( Subtract(Light.Position, Faces[AFace].Center) );

      LIntensityRatio := DotProduct(Faces[AFace].Normal, LSourceDirection);
      if LIntensityRatio < 0 then LIntensityRatio := 0;

      Faces[AFace].HLS.L := Min(Ambient + trunc(Light.Luminance * LIntensityRatio), 255);

      LColor := RGBToCol( HLStoRGB(Faces[AFace].HLS) );

      // create polygon array with pre-calculated perspective projections
      for LVertex := 0 to High(LPolygon) do
        with Vertices[ Faces[AFace].Corners[LVertex] ].ScreenPoint do begin
          LPolygon[LVertex].X := Round(X * FSuperRatio);
          LPolygon[LVertex].Y := Round(Y * FSuperRatio);
        end;

    end;

    // draw face
    if SuperSample > 0 then begin
      SuperBmp.Canvas.Pen.Color := LColor;
      SuperBmp.Canvas.Brush.Color := LColor;
      SuperBmp.Canvas.Polygon(LPolygon);
    end else begin
      OffScrBmp.Canvas.Pen.Color := LColor;
      OffScrBmp.Canvas.Brush.Color := LColor;
      OffScrBmp.Canvas.Polygon(LPolygon);
    end;

  except
    ShowMessage('Exception in World FlatFace Method');
  end;
end;

// render the face of an object using Gouraud shading
procedure T3DWorld.GouraudFace(AObject, AFace : integer);
Var
  LPolygon : TPointColorTriangle;
  LVertex : byte;
  LScreenPoint : TPointFloat;
  LTriangle : integer;
  LVertexChange : integer;
begin
  try

    with Objects[AObject] do begin

      // set up first 2 vertices of the first triangle
      for LVertex := 0 to 1 do begin
        LScreenPoint := Vertices[ Faces[AFace].Corners[LVertex] ].ScreenPoint;
        LPolygon[LVertex].X := LScreenPoint.X * FSuperRatio;
        LPolygon[LVertex].Y := LScreenPoint.Y * FSuperRatio;
        Faces[AFace].HLS.L := Vertices[Faces[AFace].Corners[LVertex]].Lum;
        LPolygon[LVertex].RGB := HLStoRGB( Faces[AFace].HLS );
      end;

      // draw all the triangles making up the polygon by walking 2 of the triangle vertices around the polygon
      for LTriangle := 0 to length(Faces[AFace].Corners) - 3 do begin
        if Odd(LTriangle) then LVertexChange := 1 else LVertexChange := 2;
        LScreenPoint := Vertices[ Faces[AFace].Corners[LTriangle + 2] ].ScreenPoint;
        LPolygon[LVertexChange].X := LScreenPoint.X * FSuperRatio;
        LPolygon[LVertexChange].Y := LScreenPoint.Y * FSuperRatio;
        Faces[AFace].HLS.L := Vertices[Faces[AFace].Corners[LTriangle + 2]].Lum;
        LPolygon[LVertexChange].RGB := HLStoRGB( Faces[AFace].HLS );
        GouraudPoly(ScreenRect.Right  shl FSSshr1-1,
                    ScreenRect.Bottom shl FSSshr1-1, LPolygon);
      end;

    end;

  except
    ShowMessage('Exception in World GouraudFace Method');
  end;
end;

// render the face of an object using Phong shading
procedure T3DWorld.PhongFace(AObject, AFace : integer);
Var
  LPolygon : TPointLightTriangle;
  LVertex : byte;
  LScreenPoint : TPointFloat;
  LTriangle : integer;
  LVertexChange : integer;
begin
  try

    with Objects[AObject] do begin

      // set up first 2 vertices of the first triangle
      for LVertex := 0 to 1 do begin
        LScreenPoint := Vertices[ Faces[AFace].Corners[LVertex] ].ScreenPoint;
        LPolygon[LVertex].X := LScreenPoint.X * FSuperRatio;
        LPolygon[LVertex].Y := LScreenPoint.Y * FSuperRatio;
        LPolygon[LVertex].N := Point3DDtoS(Vertices[Faces[AFace].Corners[LVertex]].Normal);
        LPolygon[LVertex].S := Point3DDtoS(Subtract(Light.Position,
                               Vertices[Faces[AFace].Corners[LVertex]].Point));
      end;

      // draw all the triangles making up the polygon by walking 2 of the triangle vertices around the polygon
      for LTriangle := 0 to length(Faces[AFace].Corners) - 3 do begin
        if Odd(LTriangle) then LVertexChange := 1 else LVertexChange := 2;
        LScreenPoint := Vertices[Faces[AFace].Corners[LTriangle + 2]].ScreenPoint;
        LPolygon[LVertexChange].X := LScreenPoint.X * FSuperRatio;
        LPolygon[LVertexChange].Y := LScreenPoint.Y * FSuperRatio;
        LPolygon[LVertexChange].N := Point3DDtoS(Vertices[Faces[AFace].Corners[LTriangle + 2]].Normal);
        LPolygon[LVertexChange].S := Point3DDtoS(Subtract(Light.Position,
                                     Vertices[Faces[AFace].Corners[LTriangle + 2]].Point));
        PhongPoly(ScreenRect.Right shl FSSshr1-1, ScreenRect.Bottom shl FSSshr1-1,
                  LPolygon, Faces[AFace].HLS.H, Faces[AFace].HLS.S,
                  Light.Luminance, Ambient);
      end;

    end;

  except
    ShowMessage('Exception in World PhongFace Method');
  end;
end;

// fill a traingular polygon using Gouraud shading
// interpolate vertex RGB values across triangle
procedure T3DWorld.GouraudPoly(W, H : integer ; V : TPointColorTriangle);
Var
  LX, RX, Ldx, Rdx : Single;
  Dif1, Dif1Rec, Dif2, Dif2Rec : Single;
  LRGB, RRGB, LRGBdy, RRGBdy : TRGBFixed;
  R, G, B, RGBdxR, RGBdxG, RGBdxB : integer;
  y, x, ScanStart, ScanEnd, ScanEndUnroll : integer;
  Vmax : byte;
  Right : boolean;
  Temp : TPointColor;
  VRGBF : array[0..2] of TRGBFixed;
  scanpointer : pbyte;
begin
  try

    // sort vertices by Y
    Vmax := 0;
    if V[1].Y > V[0].Y then Vmax := 1;
    if V[2].Y > V[Vmax].Y then Vmax := 2;
    if Vmax <> 2 then begin
      Temp := V[2];
      V[2] := V[Vmax];                     //           /\
      V[Vmax] := Temp;                     //          /  \ region 1
    end;                                   //         /____\
    if V[1].Y > V[0].Y then Vmax := 1      //        /    /
                       else Vmax := 0;     //       /   / region 2
    if Vmax = 0 then begin                 //      /  /
      Temp := V[1];                        //     / /
      V[1] := V[0];                        //    /
      V[0] := Temp;
    end;

    Dif1 := V[2].Y - V[0].Y;
    if Dif1 = 0 then Dif1 := 0.001; // prevent EZeroDivide
    Dif1Rec := 1 / Dif1;
    Dif2 := V[1].Y - V[0].Y;
    if Dif2 = 0 then Dif2 := 0.001;
    Dif2Rec := 1 / Dif2;

    { work out if middle point is to the left or right of the line
      connecting upper and lower points }
    if V[1].X > (V[2].X - V[0].X) * Dif2 * Dif1Rec + V[0].X then Right := True
                                                            else Right := False;

    // create fixed point rgb values
    for x := 0 to 2 do begin
      VRGBF[x].R := V[x].RGB.R shl 23;
      VRGBF[x].G := V[x].RGB.G shl 23;
      VRGBF[x].B := V[x].RGB.B shl 23;
    end;

    // calculate increments in x and colour for stepping through the lines
    if Right then begin
      Ldx := (V[2].X - V[0].X) * Dif1Rec;
      Rdx := (V[1].X - V[0].X) * Dif2Rec;
      LRGBdy.B := Trunc((VRGBF[2].B - VRGBF[0].B) * Dif1Rec);
      LRGBdy.G := Trunc((VRGBF[2].G - VRGBF[0].G) * Dif1Rec);
      LRGBdy.R := Trunc((VRGBF[2].R - VRGBF[0].R) * Dif1Rec);
      RRGBdy.B := Trunc((VRGBF[1].B - VRGBF[0].B) * Dif2Rec);
      RRGBdy.G := Trunc((VRGBF[1].G - VRGBF[0].G) * Dif2Rec);
      RRGBdy.R := Trunc((VRGBF[1].R - VRGBF[0].R) * Dif2Rec);
    end else begin
      Ldx := (V[1].X - V[0].X) * Dif2Rec;
      Rdx := (V[2].X - V[0].X) * Dif1Rec;
      RRGBdy.B := Trunc((VRGBF[2].B - VRGBF[0].B) * Dif1Rec);
      RRGBdy.G := Trunc((VRGBF[2].G - VRGBF[0].G) * Dif1Rec);
      RRGBdy.R := Trunc((VRGBF[2].R - VRGBF[0].R) * Dif1Rec);
      LRGBdy.B := Trunc((VRGBF[1].B - VRGBF[0].B) * Dif2Rec);
      LRGBdy.G := Trunc((VRGBF[1].G - VRGBF[0].G) * Dif2Rec);
      LRGBdy.R := Trunc((VRGBF[1].R - VRGBF[0].R) * Dif2Rec);
    end;

    LRGB := VRGBF[0];
    RRGB := LRGB;

    LX := V[0].X;
    RX := V[0].X;

    // fill region 1
    for y := Round(V[0].Y) to Round(V[1].Y) - 1 do begin

      // y clipping
      if y > H then Break;
      if y < 0 then begin
        LX := LX + Ldx;
        RX := RX + Rdx;
        Inc(LRGB.B, LRGBdy.B);
        Inc(LRGB.G, LRGBdy.G);
        Inc(LRGB.R, LRGBdy.R);
        Inc(RRGB.B, RRGBdy.B);
        Inc(RRGB.G, RRGBdy.G);
        Inc(RRGB.R, RRGBdy.R);
        Continue;
      end;

      // calculate increments in color for stepping through pixels
      Dif1 := RX - LX + 1;
      if Dif1 = 0 then Dif1 := 0.001;
      Dif1Rec := 1 / Dif1;
      RGBdxB := Trunc((RRGB.B - LRGB.B) * Dif1Rec);
      RGBdxG := Trunc((RRGB.G - LRGB.G) * Dif1Rec);
      RGBdxR := Trunc((RRGB.R - LRGB.R) * Dif1Rec);

      // x clipping
      ScanStart := Round(LX);
      if LX < 0 then begin
        B := LRGB.B - RGBdxB * ScanStart;
        G := LRGB.G - RGBdxG * ScanStart;
        R := LRGB.R - RGBdxR * ScanStart;
        ScanStart := 0;
      end else begin
        R := LRGB.R;
        G := LRGB.G;
        B := LRGB.B;
      end;
      if RX - 1 > W then ScanEnd := W
                    else ScanEnd := Round(RX) - 1;

      // scan the line
      scanpointer := @Scanlines[y][ScanStart];
      x := ScanStart;
      ScanEndUnroll := ScanEnd - 4;
      while x <= ScanEndUnroll do begin
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Inc(x, 4);
      end;
      while x <= ScanEnd do begin
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Inc(x);
      end;

      // increment edge x positions
      LX := LX + Ldx;
      RX := RX + Rdx;

      // increment edge colours by the y colour increments
      Inc(LRGB.B, LRGBdy.B);
      Inc(LRGB.G, LRGBdy.G);
      Inc(LRGB.R, LRGBdy.R);
      Inc(RRGB.B, RRGBdy.B);
      Inc(RRGB.G, RRGBdy.G);
      Inc(RRGB.R, RRGBdy.R);

    end;

    Dif1 := V[2].Y - V[1].Y;
    if Dif1 = 0 then Dif1 := 0.001;
    Dif1Rec := 1 / Dif1;
    // calculate new increments for region 2
    if Right then begin
      Rdx := (V[2].X - V[1].X) * Dif1Rec;
      RX := V[1].X;
      RRGBdy.B := Trunc((VRGBF[2].B - VRGBF[1].B) * Dif1Rec);
      RRGBdy.G := Trunc((VRGBF[2].G - VRGBF[1].G) * Dif1Rec);
      RRGBdy.R := Trunc((VRGBF[2].R - VRGBF[1].R) * Dif1Rec);
      RRGB := VRGBF[1];
    end else begin
      Ldx := (V[2].X - V[1].X) * Dif1Rec;
      LX := V[1].X;
      LRGBdy.B := Trunc((VRGBF[2].B - VRGBF[1].B) * Dif1Rec);
      LRGBdy.G := Trunc((VRGBF[2].G - VRGBF[1].G) * Dif1Rec);
      LRGBdy.R := Trunc((VRGBF[2].R - VRGBF[1].R) * Dif1Rec);
      LRGB := VRGBF[1];
    end;

    // fill region 2
    for y := Round(V[1].Y) to Round(V[2].Y) - 1 do begin

      // y clipping
      if y > H then Break;
      if y < 0 then begin
        LX := LX + Ldx;
        RX := RX + Rdx;
        Inc(LRGB.B, LRGBdy.B);
        Inc(LRGB.G, LRGBdy.G);
        Inc(LRGB.R, LRGBdy.R);
        Inc(RRGB.B, RRGBdy.B);
        Inc(RRGB.G, RRGBdy.G);
        Inc(RRGB.R, RRGBdy.R);
        Continue;
      end;

      Dif1 := RX - LX + 1;
      if Dif1 = 0 then Dif1 := 0.001;
      Dif1Rec := 1 / Dif1;
      RGBdxB := Trunc((RRGB.B - LRGB.B) * Dif1Rec);
      RGBdxG := Trunc((RRGB.G - LRGB.G) * Dif1Rec);
      RGBdxR := Trunc((RRGB.R - LRGB.R) * Dif1Rec);

      // x clipping
      ScanStart := Round(LX);
      if LX < 0 then begin
        B := LRGB.B - RGBdxB * ScanStart;
        G := LRGB.G - RGBdxG * ScanStart;
        R := LRGB.R - RGBdxR * ScanStart;
        ScanStart := 0;
      end else begin
        R := LRGB.R;
        G := LRGB.G;
        B := LRGB.B;
      end;
      if RX - 1 > W then ScanEnd := W
                    else ScanEnd := Round(RX) - 1;

      // scan the line
      scanpointer := @Scanlines[y][ScanStart];
      x := ScanStart;
      ScanEndUnroll := ScanEnd - 4;
      while x <= ScanEndUnroll do begin
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Inc(x, 4);
      end;
      while x <= ScanEnd do begin
        Scanpointer^ := B shr 23;
        Inc(scanpointer);
        Scanpointer^ := G shr 23;
        Inc(scanpointer);
        Scanpointer^ := R shr 23;
        Inc(scanpointer,2);
        Inc(B, RGBdxB);
        Inc(G, RGBdxG);
        Inc(R, RGBdxR);
        Inc(x);
      end;

      LX := LX + Ldx;
      RX := RX + Rdx;

      Inc(LRGB.B, LRGBdy.B);
      Inc(LRGB.G, LRGBdy.G);
      Inc(LRGB.R, LRGBdy.R);
      Inc(RRGB.B, RRGBdy.B);
      Inc(RRGB.G, RRGBdy.G);
      Inc(RRGB.R, RRGBdy.R);

    end;

  except
    ShowMessage('Exception in GouraudPoly Method');
  end;
end;

// fill a traingular polygon using phong shading
// interpolate vertex normals across triangle
procedure T3DWorld.PhongPoly(W, H : integer ; V : TPointLightTriangle ;
                             Hue, Sat, Lum, Amb : byte);
Var
  LX, RX, Ldx, Rdx : Single;
  Dif1, Dif1Rec, Dif2, Dif2Rec : Single;
  LN, RN, N, Ndx, LNdy, RNdy, NNorm : TPoint3DSingle;
  LS, RS, S, Sdx, LSdy, RSdy, SNorm : TPoint3DSingle;
  y, x, ScanStart, ScanEnd : integer;
  Vmax : byte;
  Right : boolean;
  Temp : TPointLight;
  IR : Single; // intensity ratio
  HLS : THLS;
begin
  try

    // sort vertices by Y
    Vmax := 0;
    if V[1].Y > V[0].Y then Vmax := 1;
    if V[2].Y > V[Vmax].Y then Vmax := 2;
    if Vmax <> 2 then begin
      Temp := V[2];
      V[2] := V[Vmax];                     //           /\
      V[Vmax] := Temp;                     //          /  \ region 1
    end;                                   //         /____\
    if V[1].Y > V[0].Y then Vmax := 1      //        /    /
                       else Vmax := 0;     //       /   / region 2
    if Vmax = 0 then begin                 //      /  /
      Temp := V[1];                        //     / /
      V[1] := V[0];                        //    /
      V[0] := Temp;
    end;

    Dif1 := V[2].Y - V[0].Y;
    if Dif1 = 0 then Dif1 := 0.001; // prevent EZeroDivide
    Dif1Rec := 1 / Dif1;
    Dif2 := V[1].Y - V[0].Y;
    if Dif2 = 0 then Dif2 := 0.001;
    Dif2Rec := 1 / Dif2;

    { work out if middle point is to the left or right of the line
      connecting upper and lower points }
    if V[1].X > (V[2].X - V[0].X) * Dif2 * Dif1Rec + V[0].X then Right := True
                                                            else Right := False;

    // calculate increments in x, normal and source for stepping through the lines
    if Right then begin
      Ldx := (V[2].X - V[0].X) * Dif1Rec;
      Rdx := (V[1].X - V[0].X) * Dif2Rec;
      LNdy.X := (V[2].N.X - V[0].N.X) * Dif1Rec;
      LNdy.Y := (V[2].N.Y - V[0].N.Y) * Dif1Rec;
      LNdy.Z := (V[2].N.Z - V[0].N.Z) * Dif1Rec;
      RNdy.X := (V[1].N.X - V[0].N.X) * Dif2Rec;
      RNdy.Y := (V[1].N.Y - V[0].N.Y) * Dif2Rec;
      RNdy.Z := (V[1].N.Z - V[0].N.Z) * Dif2Rec;
      LSdy.X := (V[2].S.X - V[0].S.X) * Dif1Rec;
      LSdy.Y := (V[2].S.Y - V[0].S.Y) * Dif1Rec;
      LSdy.Z := (V[2].S.Z - V[0].S.Z) * Dif1Rec;
      RSdy.X := (V[1].S.X - V[0].S.X) * Dif2Rec;
      RSdy.Y := (V[1].S.Y - V[0].S.Y) * Dif2Rec;
      RSdy.Z := (V[1].S.Z - V[0].S.Z) * Dif2Rec;
    end else begin
      Ldx := (V[1].X - V[0].X) * Dif2Rec;
      Rdx := (V[2].X - V[0].X) * Dif1Rec;
      RNdy.X := (V[2].N.X - V[0].N.X) * Dif1Rec;
      RNdy.Y := (V[2].N.Y - V[0].N.Y) * Dif1Rec;
      RNdy.Z := (V[2].N.Z - V[0].N.Z) * Dif1Rec;
      LNdy.X := (V[1].N.X - V[0].N.X) * Dif2Rec;
      LNdy.Y := (V[1].N.Y - V[0].N.Y) * Dif2Rec;
      LNdy.Z := (V[1].N.Z - V[0].N.Z) * Dif2Rec;
      RSdy.X := (V[2].S.X - V[0].S.X) * Dif1Rec;
      RSdy.Y := (V[2].S.Y - V[0].S.Y) * Dif1Rec;
      RSdy.Z := (V[2].S.Z - V[0].S.Z) * Dif1Rec;
      LSdy.X := (V[1].S.X - V[0].S.X) * Dif2Rec;
      LSdy.Y := (V[1].S.Y - V[0].S.Y) * Dif2Rec;
      LSdy.Z := (V[1].S.Z - V[0].S.Z) * Dif2Rec;
    end;

    LN := V[0].N;
    RN := LN;
    LS := V[0].S;
    RS := LS;

    LX := V[0].X;
    RX := LX;

    HLS.H := Hue;
    HLS.S := Sat;

    // fill region 1
    for y := Round(V[0].Y) to Round(V[1].Y) - 1 do begin

      // y clipping
      if y > H then Break;
      if y < 0 then begin
        LX := LX + Ldx;
        RX := RX + Rdx;
        LN.X := LN.X + LNdy.X;
        LN.Y := LN.Y + LNdy.Y;
        LN.Z := LN.Z + LNdy.Z;
        RN.X := RN.X + RNdy.X;
        RN.Y := RN.Y + RNdy.Y;
        RN.Z := RN.Z + RNdy.Z;
        LS.X := LS.X + LSdy.X;
        LS.Y := LS.Y + LSdy.Y;
        LS.Z := LS.Z + LSdy.Z;
        RS.X := RS.X + RSdy.X;
        RS.Y := RS.Y + RSdy.Y;
        RS.Z := RS.Z + RSdy.Z;
        Continue;
      end;

      // calculate increments in normal for stepping through pixels
      Dif1 := RX - LX + 1;
      if Dif1 = 0 then Dif1 := 0.001;
      Dif1Rec := 1 / Dif1;
      Ndx.X := (RN.X - LN.X) * Dif1Rec;
      Ndx.Y := (RN.Y - LN.Y) * Dif1Rec;
      Ndx.Z := (RN.Z - LN.Z) * Dif1Rec;
      Sdx.X := (RS.X - LS.X) * Dif1Rec;
      Sdx.Y := (RS.Y - LS.Y) * Dif1Rec;
      Sdx.Z := (RS.Z - LS.Z) * Dif1Rec;

      // x clipping
      ScanStart := Round(LX);
      if LX < 0 then begin
        N.X := LN.X - Ndx.X * ScanStart;
        N.Y := LN.Y - Ndx.Y * ScanStart;
        N.Z := LN.Z - Ndx.Z * ScanStart;
        S.X := LS.X - Sdx.X * ScanStart;
        S.Y := LS.Y - Sdx.Y * ScanStart;
        S.Z := LS.Z - Sdx.Z * ScanStart;
        ScanStart := 0;
      end else begin
        N := LN;
        S := LS;
      end;
      if RX - 1 > W then ScanEnd := W else ScanEnd := Round(RX) - 1;

      // scan the line
      for x := ScanStart to ScanEnd do begin
        NNorm := UnitVectorS(N);
        SNorm := UnitVectorS(S);
        IR := DotProductS(NNorm, SNorm);
        if IR < 0 then HLS.L := Amb else HLS.L := Round(Amb + (Lum * IR));
        //if HLS.L > 255 then HLS.L := 255;
        Scanlines[y][x] := HLStoRGB(HLS);
        N.X := N.X + Ndx.X;
        N.Y := N.Y + Ndx.Y;
        N.Z := N.Z + Ndx.Z;
        S.X := S.X + Sdx.X;
        S.Y := S.Y + Sdx.Y;
        S.Z := S.Z + Sdx.Z;
      end;
      // increment edge x positions
      LX := LX + Ldx;
      RX := RX + Rdx;

      // increment edge normals by the y colour increments
      LN.X := LN.X + LNdy.X;
      LN.Y := LN.Y + LNdy.Y;
      LN.Z := LN.Z + LNdy.Z;
      RN.X := RN.X + RNdy.X;
      RN.Y := RN.Y + RNdy.Y;
      RN.Z := RN.Z + RNdy.Z;
      LS.X := LS.X + LSdy.X;
      LS.Y := LS.Y + LSdy.Y;
      LS.Z := LS.Z + LSdy.Z;
      RS.X := RS.X + RSdy.X;
      RS.Y := RS.Y + RSdy.Y;
      RS.Z := RS.Z + RSdy.Z;

    end;

    Dif1 := V[2].Y - V[1].Y;
    if Dif1 = 0 then Dif1 := 0.001;
    Dif1Rec := 1 / Dif1;
    // calculate new increments for region 2
    if Right then begin
      Rdx := (V[2].X - V[1].X) * Dif1Rec;
      RX := V[1].X;
      RNdy.X := (V[2].N.X - V[1].N.X) * Dif1Rec;
      RNdy.Y := (V[2].N.Y - V[1].N.Y) * Dif1Rec;
      RNdy.Z := (V[2].N.Z - V[1].N.Z) * Dif1Rec;
      RSdy.X := (V[2].S.X - V[1].S.X) * Dif1Rec;
      RSdy.Y := (V[2].S.Y - V[1].S.Y) * Dif1Rec;
      RSdy.Z := (V[2].S.Z - V[1].S.Z) * Dif1Rec;
      RN := V[1].N;
      RS := V[1].S;
    end else begin
      Ldx := (V[2].X - V[1].X) * Dif1Rec;
      LX := V[1].X;
      LNdy.X := (V[2].N.X - V[1].N.X) * Dif1Rec;
      LNdy.Y := (V[2].N.Y - V[1].N.Y) * Dif1Rec;
      LNdy.Z := (V[2].N.Z - V[1].N.Z) * Dif1Rec;
      LSdy.X := (V[2].S.X - V[1].S.X) * Dif1Rec;
      LSdy.Y := (V[2].S.Y - V[1].S.Y) * Dif1Rec;
      LSdy.Z := (V[2].S.Z - V[1].S.Z) * Dif1Rec;
      LN := V[1].N;
      LS := V[1].S;
    end;

    // fill region 2
    for y := Round(V[1].Y) to Round(V[2].Y) - 1 do begin

      // y clipping
      if y > H then Break;
      if y < 0 then begin
        LX := LX + Ldx;
        RX := RX + Rdx;
        LN.X := LN.X + LNdy.X;
        LN.Y := LN.Y + LNdy.Y;
        LN.Z := LN.Z + LNdy.Z;
        RN.X := RN.X + RNdy.X;
        RN.Y := RN.Y + RNdy.Y;
        RN.Z := RN.Z + RNdy.Z;
        LS.X := LS.X + LSdy.X;
        LS.Y := LS.Y + LSdy.Y;
        LS.Z := LS.Z + LSdy.Z;
        RS.X := RS.X + RSdy.X;
        RS.Y := RS.Y + RSdy.Y;
        RS.Z := RS.Z + RSdy.Z;
        Continue;
      end;

      Dif1 := RX - LX + 1;
      if Dif1 = 0 then Dif1 := 0.001;
      Dif1Rec := 1 / Dif1;
      Ndx.X := (RN.X - LN.X) * Dif1Rec;
      Ndx.Y := (RN.Y - LN.Y) * Dif1Rec;
      Ndx.Z := (RN.Z - LN.Z) * Dif1Rec;
      Sdx.X := (RS.X - LS.X) * Dif1Rec;
      Sdx.Y := (RS.Y - LS.Y) * Dif1Rec;
      Sdx.Z := (RS.Z - LS.Z) * Dif1Rec;

      // x clipping
      ScanStart := Round(LX);
      if LX < 0 then begin
        N.X := LN.X - Ndx.X * ScanStart;
        N.Y := LN.Y - Ndx.Y * ScanStart;
        N.Z := LN.Z - Ndx.Z * ScanStart;
        S.X := LS.X - Sdx.X * ScanStart;
        S.Y := LS.Y - Sdx.Y * ScanStart;
        S.Z := LS.Z - Sdx.Z * ScanStart;
        ScanStart := 0;
      end else begin
        N := LN;
        S := LS;
      end;
      if RX - 1 > W then ScanEnd := W else ScanEnd := Round(RX) - 1;

      // scan the line
      for x := ScanStart to ScanEnd do begin
        NNorm := UnitVectorS(N);
        SNorm := UnitVectorS(S);
        IR := DotProductS(NNorm, SNorm);
        if IR < 0 then HLS.L := Amb else HLS.L := Round(Amb + (Lum * IR));
        //if HLS.L > 255 then HLS.L := 255;
        Scanlines[y][x] := HLStoRGB(HLS);
        N.X := N.X + Ndx.X;
        N.Y := N.Y + Ndx.Y;
        N.Z := N.Z + Ndx.Z;
        S.X := S.X + Sdx.X;
        S.Y := S.Y + Sdx.Y;
        S.Z := S.Z + Sdx.Z;
      end;

      LX := LX + Ldx;
      RX := RX + Rdx;

      LN.X := LN.X + LNdy.X;
      LN.Y := LN.Y + LNdy.Y;
      LN.Z := LN.Z + LNdy.Z;
      RN.X := RN.X + RNdy.X;
      RN.Y := RN.Y + RNdy.Y;
      RN.Z := RN.Z + RNdy.Z;
      LS.X := LS.X + LSdy.X;
      LS.Y := LS.Y + LSdy.Y;
      LS.Z := LS.Z + LSdy.Z;
      RS.X := RS.X + RSdy.X;
      RS.Y := RS.Y + RSdy.Y;
      RS.Z := RS.Z + RSdy.Z;

    end;

  except
    ShowMessage('Exception in World PhongPoly Method');
  end;
end;

// anti-aliased line. uses fixed point arithmetic
procedure T3DWorld.WuLine(x1, y1, x2, y2, W, H : integer ; R, G, B : byte);
var
  deltax, deltay, X, Y, start, finish : integer;
  LM, LR : integer;
  dxi, dyi, dydxi : integer;
begin
  deltax := abs(x2 - x1); // Calculate deltax and deltay for initialisation
  deltay := abs(y2 - y1);
  if (deltax = 0) or (deltay = 0) then begin
    if SuperSample > 0 then begin
      SuperBmp.Canvas.Pen.Color := (B shl 16) + (G shl 8) + R;
      SuperBmp.Canvas.MoveTo(x1, y1);
      SuperBmp.Canvas.LineTo(x2, y2);
    end else begin
      OffScrBmp.Canvas.Pen.Color := (B shl 16) + (G shl 8) + R;
      OffScrBmp.Canvas.MoveTo(x1, y1);
      OffScrBmp.Canvas.LineTo(x2, y2);
    end;
    exit;
  end;
  if deltax > deltay then begin // horizontal or vertical
    if y2 > y1 then // determine rise and run
      dydxi := -deltay shl 16 div deltax
    else
      dydxi := deltay shl 16 div deltax;
    if x2 < x1 then begin
      start := x2; // right to left
      finish := x1;
      dyi := y2 shl 16;
    end else begin
      start := x1; // left to right
      finish := x2;
      dyi := y1 shl 16;
      dydxi := -dydxi; // inverse slope
    end;
    if finish >= W then finish := W - 1;
    for X := start to finish do begin
      Y := dyi shr 16;
      if (X < 0) or (Y < 0) or (Y > H-2) then begin
        Inc(dyi, dydxi);
        Continue;
      end;
      LM := dyi - Y shl 16; // fractional part of dyi - in fixed-point
      LR := 65536 - LM;
      Scanlines[Y][X].B := (B*LR + Scanlines[Y][X].B*LM) shr 16;
      Scanlines[Y][X].G := (G*LR + Scanlines[Y][X].G*LM) shr 16;
      Scanlines[Y][X].R := (R*LR + Scanlines[Y][X].R*LM) shr 16;
      Inc(Y);
      Scanlines[Y][X].B := (B*LM + Scanlines[Y][X].B*LR) shr 16;
      Scanlines[Y][X].G := (G*LM + Scanlines[Y][X].G*LR) shr 16;
      Scanlines[Y][X].R := (R*LM + Scanlines[Y][X].R*LR) shr 16;
      Inc(dyi, dydxi); // next point
    end;
  end else begin
    if x2 > x1 then // determine rise and run
      dydxi := -deltax shl 16 div deltay
    else
      dydxi := deltax shl 16 div deltay;
    if y2 < y1 then begin
      start := y2; // right to left
      finish := y1;
      dxi := x2 shl 16;
    end else begin
      start := y1; // left to right
      finish := y2;
      dxi := x1 shl 16;
      dydxi := -dydxi; // inverse slope
    end;
    if finish >= H then finish := H - 1;
    for Y := start to finish do begin
      X := dxi shr 16;
      if (Y < 0) or (X < 0) or (X > W-2) then begin
        Inc(dxi, dydxi);
        Continue;
      end;
      LM := dxi - X shl 16;
      LR := 65536 - LM;
      Scanlines[Y][X].B := (B*LR + Scanlines[Y][X].B*LM) shr 16;
      Scanlines[Y][X].G := (G*LR + Scanlines[Y][X].G*LM) shr 16;
      Scanlines[Y][X].R := (R*LR + Scanlines[Y][X].R*LM) shr 16;
      Inc(X);
      Scanlines[Y][X].B := (B*LM + Scanlines[Y][X].B*LR) shr 16;
      Scanlines[Y][X].G := (G*LM + Scanlines[Y][X].G*LR) shr 16;
      Scanlines[Y][X].R := (R*LM + Scanlines[Y][X].R*LR) shr 16;
      Inc(dxi, dydxi); // next point
    end;
  end;
end;

// shrink the super-sampling bitmap onto the off screen bitmap
// using pixel averaging
procedure T3DWorld.SuperSampleBitmap;
Var
  Lx, Ly, LxS, LyS, LySend, LxSend, LyB, LxB : integer;
  P1, P2, P3, P4, P5, P6, P7, P8, P9, ScanPointerOut : pByte;
  PA, PB, PC, PD, PE, PF, PG : pByte;
  LSumR, LSumG, LSumB, LBoxPixels : integer;
  W, H, WM, HM, WL : integer;
begin
  try

    W := ScreenRect.Right;
    H := ScreenRect.Bottom;
    WM := W - 1; HM := H - 1;
    WL := W - 3;
    LyS := 0;
    if FSuperRatio = 2 then for Ly := 0 to WM do begin
      ScanPointerOut := @ScanlinesOSBmp[Ly][0];
      P1 := @Scanlines[LyS][0]; Integer(P2) := Integer(P1) + 4;
      P3 := @Scanlines[LyS+1][0]; Integer(P4) := Integer(P3) + 4;
      Lx := 0;
      // set output pixel to the average of the 2X2 box of input pixels
      while Lx < WL do begin // loop unrolled by 4
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // blue
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // green
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // red
        Inc(P1,6); Inc(P2,6); Inc(P3,6); Inc(P4,6); Inc(ScanPointerOut, 2);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // blue
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // green
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // red
        Inc(P1,6); Inc(P2,6); Inc(P3,6); Inc(P4,6); Inc(ScanPointerOut, 2);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // blue
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // green
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // red
        Inc(P1,6); Inc(P2,6); Inc(P3,6); Inc(P4,6); Inc(ScanPointerOut, 2);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // blue
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // green
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // red
        Inc(P1,6); Inc(P2,6); Inc(P3,6); Inc(P4,6); Inc(ScanPointerOut, 2);
        Inc(Lx, 4);
      end;
      while Lx < W do begin
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // blue
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // green
        Inc(P1); Inc(P2); Inc(P3); Inc(P4); Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^ + P2^ + P3^ + P4^) shr 2; // red
        Inc(P1,6); Inc(P2,6); Inc(P3,6); Inc(P4,6); Inc(ScanPointerOut, 2);
        Inc(Lx);
      end;
      Inc(LyS, 2);
    end else if FSuperRatio = 4 then for Ly := 0 to HM do begin
      ScanPointerOut := @ScanlinesOSBmp[Ly][0];
      P1 := @Scanlines[LyS][0]; Integer(P2) := Integer(P1) + 4;
      Integer(P3) := Integer(P2) + 4; Integer(P4) := Integer(P3) + 4;
      P5 := @Scanlines[LyS+1][0]; Integer(P6) := Integer(P5) + 4;
      Integer(P7) := Integer(P6) + 4; Integer(P8) := Integer(P7) + 4;
      P9 := @Scanlines[LyS+2][0]; Integer(PA) := Integer(P9) + 4;
      Integer(PB) := Integer(PA) + 4; Integer(PC) := Integer(PB) + 4;
      PD := @Scanlines[LyS+3][0]; Integer(PE) := Integer(PD) + 4;
      Integer(PF) := Integer(PE) + 4; Integer(PG) := Integer(PF) + 4;
      for Lx := 0 to WM do begin
        // set output pixel to the average of the 4X4 box of input pixels
        ScanPointerOut^ := (P1^+P2^+P3^+P4^+P5^+P6^+P7^+P8^+P9^+PA^+PB^+PC^+PD^+PE^+PF^+PG^) shr 4;
        Inc(P1);Inc(P2);Inc(P3);Inc(P4);Inc(P5);Inc(P6);Inc(P7);Inc(P8);
        Inc(P9);Inc(PA);Inc(PB);Inc(PC);Inc(PD);Inc(PE);Inc(PF);Inc(PG);
        Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^+P2^+P3^+P4^+P5^+P6^+P7^+P8^+P9^+PA^+PB^+PC^+PD^+PE^+PF^+PG^) shr 4;
        Inc(P1);Inc(P2);Inc(P3);Inc(P4);Inc(P5);Inc(P6);Inc(P7);Inc(P8);
        Inc(P9);Inc(PA);Inc(PB);Inc(PC);Inc(PD);Inc(PE);Inc(PF);Inc(PG);
        Inc(ScanPointerOut);
        ScanPointerOut^ := (P1^+P2^+P3^+P4^+P5^+P6^+P7^+P8^+P9^+PA^+PB^+PC^+PD^+PE^+PF^+PG^) shr 4;
        Inc(P1,14);Inc(P2,14);Inc(P3,14);Inc(P4,14);Inc(P5,14);Inc(P6,14);Inc(P7,14);Inc(P8,14);
        Inc(P9,14);Inc(PA,14);Inc(PB,14);Inc(PC,14);Inc(PD,14);Inc(PE,14);Inc(PF,14);Inc(PG,14);
        Inc(ScanPointerOut, 2);
      end;
      Inc(LyS, 4);
    end else begin // shrink by any ratio > 4
      for Ly := 0 to HM do begin
        LySend := LyS + FSuperRatio - 1;
        ScanPointerOut := @ScanlinesOSBmp[Ly][0];
        LxS := 0;
        for Lx := 0 to WM do begin
          LxSend := LxS + FSuperRatio - 1;
          LSumR := 0; LSumG := 0; LSumB := 0;
          // loop through super-sampling box of pixels in input bitmap and sum them
          for LyB := LyS to LySend do begin
            P1 := @Scanlines[LyB][LxS];
            for LxB := LxS to LxSend do begin
              Inc(LSumB, P1^); Inc(P1);
              Inc(LSumG, P1^); Inc(P1);
              Inc(LSumR, P1^); Inc(P1, 2);
            end;
          end;
          // set output bitmap pixel to the average of the pixels in the box
          ScanPointerOut^ := LSumB shr FSuperSample;
          Inc(ScanPointerOut);
          ScanPointerOut^ := LSumG shr FSuperSample;
          Inc(ScanPointerOut);
          ScanPointerOut^ := LSumR shr FSuperSample;
          Inc(ScanPointerOut, 2);
          Inc(LxS, FSuperRatio);
        end;
        Inc(LyS, FSuperRatio);
      end;
    end;

  except
    ShowMessage('Exception in World SuperSampleBitmap method');
  end;
end;

procedure T3DWorld.ClearBitmap;
begin
  try

    OffScrBmp.Canvas.Brush.Color := BackgroundColor;
    OffScrBmp.Canvas.FillRect(ScreenRect);

  except
    ShowMessage('Exception in World ClearBitmap method');
  end;
end;

// save the world data to a memory stream
procedure T3DWorld.SaveToStream(AStream : TMemoryStream);
Var
  LRenderMethod : byte;
  LObject, LLen : integer;
begin
  try

    LRenderMethod := 0;
    if FWireFrame then LRenderMethod := 1 else
    if FPhong then LRenderMethod := 2;

    with AStream do begin

      Write(FScreenRect, SizeOf(TRect));
      Write(LRenderMethod, 1);
      Write(FBackgroundColor, SizeOf(TColor));
      Write(FSuperSample, 1);
      Write(FBackfaceRemoval, SizeOf(Boolean));
      Write(FAmbient, 1);
      Write(Light, SizeOf(TLight));

      Camera.SaveToStream(AStream);

      LLen := Length(Objects);
      Write(LLen, SizeOf(Integer));
      if LLen > 0 then for LObject := 0 to LLen - 1 do
        Objects[LObject].SaveToStream(AStream);

    end;

  except
    ShowMessage('Exception in World SaveToStream method');
  end;
end;

// load the world data from a memory stream
procedure T3DWorld.LoadFromStream(AStream : TMemoryStream);
Var
  LRenderMethod : byte;
  LObject, LLen : integer;
begin
  try

    with AStream do begin

      Read(FScreenRect, SizeOf(TRect));
      ScreenRect := FScreenRect; // set
      Read(LRenderMethod, 1);
      Phong := False; WireFrame := False;
      if LRenderMethod = 1 then WireFrame := True else
      if LRenderMethod = 2 then Phong := True;
      Read(FBackgroundColor, SizeOf(TColor));
      Read(FSuperSample, 1);
      SuperSample := FSuperSample;
      Read(FBackfaceRemoval, SizeOf(Boolean));
      Read(FAmbient, 1);
      Read(Light, SizeOf(TLight));

      Camera.LoadFromStream(AStream);

      Read(LLen, SizeOf(Integer));
      if LLen < length(Objects) then // destroy objects
        for LObject := LLen to High(Objects) do Objects[LObject].Free;
      SetLength(Objects, LLen);
      // load objects
      if LLen > 0 then for LObject := 0 to LLen - 1 do begin
        Objects[LObject] := T3DModel.Create;
        Objects[LObject].LoadFromStream(AStream);
      end;

    end;

  except
    ShowMessage('Exception in World LoadFromStream method');
  end;
end;

// save the model data to a file
procedure T3DWorld.SaveToFile(APath : String);
Var
  LStream, LStreamOut : TMemoryStream;
begin
  LStream := TMemoryStream.Create;
  LStream.Seek(0, 0);
  SaveToStream(LStream);
  LStreamOut := TMemoryStream.Create;
  PackStream(LStream, LStreamOut, clMax);
  LStream.Free;
  LStreamOut.SaveToFile(APath);
  LStreamOut.Free;
end;

// load the model data from a file
procedure T3DWorld.LoadFromFile(APath : String);
Var
  LStream, LStreamPacked : TMemoryStream;
begin
  LStreamPacked := TMemoryStream.Create;
  LStreamPacked.LoadFromFile(APath);
  LStream := TMemoryStream.Create;
  ExpandStream(LStreamPacked, LStream);
  LStreamPacked.Free;
  LStream.Seek(0, 0);
  LoadFromStream(LStream);
  LStream.Free;
end;

end.
